# coding=utf-8
# Copyright 2018 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A sum tree data structure.

Used for prioritized experience replay. See prioritized_replay_buffer.py
and Schaul et al. (2015).
"""
from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import math
import random

import numpy as np


class SumTree(object):
  """A sum tree data structure for storing replay priorities.

  A sum tree is a complete binary tree whose leaves contain values called
  priorities. Internal nodes maintain the sum of the priorities of all leaf
  nodes in their subtree.

  For capacity = 4, the tree may look like this:

               +---+
               |2.5|
               +-+-+
                 |
         +-------+--------+
         |                |
       +-+-+            +-+-+
       |1.5|            |1.0|
       +-+-+            +-+-+
         |                |
    +----+----+      +----+----+
    |         |      |         |
  +-+-+     +-+-+  +-+-+     +-+-+
  |0.5|     |1.0|  |0.5|     |0.5|
  +---+     +---+  +---+     +---+

  This is stored in a list of numpy arrays:
  self.nodes = [ [2.5], [1.5, 1], [0.5, 1, 0.5, 0.5] ]

  For conciseness, we allocate arrays as powers of two, and pad the excess
  elements with zero values.

  This is similar to the usual array-based representation of a complete binary
  tree, but is a little more user-friendly.
  """

  def __init__(self, capacity):
    """Creates the sum tree data structure for the given replay capacity.

    Args:
      capacity: int, the maximum number of elements that can be stored in this
        data structure.

    Raises:
      ValueError: If requested capacity is not positive.
    """
    assert isinstance(capacity, int)
    if capacity <= 0:
      raise ValueError('Sum tree capacity should be positive. Got: {}'.
                       format(capacity))

    self.nodes = []
    tree_depth = int(math.ceil(np.log2(capacity)))
    level_size = 1
    for _ in range(tree_depth + 1):
      nodes_at_this_depth = np.zeros(level_size)
      self.nodes.append(nodes_at_this_depth)

      level_size *= 2

    self.max_recorded_priority = 1.0

  def _total_priority(self):
    """Returns the sum of all priorities stored in this sum tree.

    Returns:
      float, sum of priorities stored in this sum tree.
    """
    return self.nodes[0][0]

  def sample(self, query_value=None):
    """Samples an element from the sum tree.

    Each element has probability p_i / sum_j p_j of being picked, where p_i is
    the (positive) value associated with node i (possibly unnormalized).

    Args:
      query_value: float in [0, 1], used as the random value to select a
      sample. If None, will select one randomly in [0, 1).

    Returns:
      int, a random element from the sum tree.

    Raises:
      Exception: If the sum tree is empty (i.e. its node values sum to 0), or if
        the supplied query_value is larger than the total sum.
    """
    if self._total_priority() == 0.0:
      raise Exception('Cannot sample from an empty sum tree.')

    if query_value and (query_value < 0. or query_value > 1.):
      raise ValueError('query_value must be in [0, 1].')

    # Sample a value in range [0, R), where R is the value stored at the root.
    query_value = random.random() if query_value is None else query_value
    query_value *= self._total_priority()

    # Now traverse the sum tree.
    node_index = 0
    for nodes_at_this_depth in self.nodes[1:]:
      # Compute children of previous depth's node.
      left_child = node_index * 2

      left_sum = nodes_at_this_depth[left_child]
      # Each subtree describes a range [0, a), where a is its value.
      if query_value < left_sum:  # Recurse into left subtree.
        node_index = left_child
      else:  # Recurse into right subtree.
        node_index = left_child + 1
        # Adjust query to be relative to right subtree.
        query_value -= left_sum

    return node_index

  def stratified_sample(self, batch_size):
    """Performs stratified sampling using the sum tree.

    Let R be the value at the root (total value of sum tree). This method will
    divide [0, R) into batch_size segments, pick a random number from each of
    those segments, and use that random number to sample from the sum_tree. This
    is as specified in Schaul et al. (2015).

    Args:
      batch_size: int, the number of strata to use.
    Returns:
      list of batch_size elements sampled from the sum tree.

    Raises:
      Exception: If the sum tree is empty (i.e. its node values sum to 0).
    """
    if self._total_priority() == 0.0:
      raise Exception('Cannot sample from an empty sum tree.')

    bounds = np.linspace(0., 1., batch_size + 1)
    assert len(bounds) == batch_size + 1
    segments = [(bounds[i], bounds[i+1]) for i in range(batch_size)]
    query_values = [random.uniform(x[0], x[1]) for x in segments]
    return [self.sample(query_value=x) for x in query_values]

  def get(self, node_index):
    """Returns the value of the leaf node corresponding to the index.

    Args:
      node_index: The index of the leaf node.
    Returns:
      The value of the leaf node.
    """
    return self.nodes[-1][node_index]

  def set(self, node_index, value):
    """Sets the value of a leaf node and updates internal nodes accordingly.

    This operation takes O(log(capacity)).
    Args:
      node_index: int, the index of the leaf node to be updated.
      value: float, the value which we assign to the node. This value must be
        nonnegative. Setting value = 0 will cause the element to never be
        sampled.

    Raises:
      ValueError: If the given value is negative.
    """
    if value < 0.0:
      raise ValueError('Sum tree values should be nonnegative. Got {}'.
                       format(value))
    self.max_recorded_priority = max(value, self.max_recorded_priority)

    delta_value = value - self.nodes[-1][node_index]

    # Now traverse back the tree, adjusting all sums along the way.
    for nodes_at_this_depth in reversed(self.nodes):
      # Note: Adding a delta leads to some tolerable numerical inaccuracies.
      nodes_at_this_depth[node_index] += delta_value
      node_index //= 2

    assert node_index == 0, ('Sum tree traversal failed, final node index '
                             'is not 0.')



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.pybind11.pyspiel."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
from absl.testing import absltest
import six

from open_spiel.python import policy
import pyspiel

# Specify game names in alphabetical order, to make the test easier to read.
EXPECTED_GAMES = set([
    "backgammon",
    "battleship",
    "blackjack",
    "blotto",
    "breakthrough",
    "bridge",
    "bridge_uncontested_bidding",
    "catch",
    "chess",
    "cliff_walking",
    "clobber",
    "coin_game",
    "connect_four",
    "coop_box_pushing",
    "coop_to_1p",
    "coordinated_mp",
    "cursor_go",
    "deep_sea",
    "efg_game",
    "first_sealed_auction",
    "gin_rummy",
    "go",
    "goofspiel",
    "havannah",
    "hex",
    "hearts",
    "kuhn_poker",
    "laser_tag",
    "lewis_signaling",
    "leduc_poker",
    "liars_dice",
    "markov_soccer",
    "matching_pennies_3p",
    "matrix_cd",
    "matrix_coordination",
    "matrix_mp",
    "matrix_pd",
    "matrix_rps",
    "matrix_rpsw",
    "matrix_sh",
    "matrix_shapleys_game",
    "misere",
    "negotiation",
    "nfg_game",
    "normal_form_extensive_game",
    "oh_hell",
    "oshi_zumo",
    "othello",
    "oware",
    "pentago",
    "phantom_ttt",
    "pig",
    "quoridor",
    "repeated_game",
    "sheriff",
    "skat",
    "start_at",
    "solitaire",
    "stones_and_gems",
    "tarok",
    "tic_tac_toe",
    "tiny_bridge_2p",
    "tiny_bridge_4p",
    "tiny_hanabi",
    "trade_comm",
    "turn_based_simultaneous_game",
    "y",
])


class PyspielTest(absltest.TestCase):

  def test_registered_names(self):
    game_names = pyspiel.registered_names()

    expected = EXPECTED_GAMES
    if os.environ.get("BUILD_WITH_HANABI", "OFF") == "ON":
      expected.add("hanabi")
    if os.environ.get("BUILD_WITH_ACPC", "OFF") == "ON":
      expected.add("universal_poker")
    expected = sorted(list(expected))
    self.assertCountEqual(game_names, expected)

  def test_no_mandatory_parameters(self):
    # Games with mandatory parameters will be skipped by several standard
    # tests. Mandatory parameters should therefore be avoided if at all
    # possible. We make a list of such games here in order to make implementors
    # think twice about adding mandatory parameters.
    def has_mandatory_params(game):
      return any(param.is_mandatory()
                 for param in game.parameter_specification.values())

    games_with_mandatory_parameters = [
        game.short_name
        for game in pyspiel.registered_games()
        if has_mandatory_params(game)
    ]
    expected = [
        # Mandatory parameters prevent various sorts of automated testing.
        # Only add games here if there is no sensible default for a parameter.
        "misere",
        "turn_based_simultaneous_game",
        "normal_form_extensive_game",
        "repeated_game",
        "start_at",
    ]
    self.assertCountEqual(games_with_mandatory_parameters, expected)

  def test_registered_game_attributes(self):
    games = {game.short_name: game for game in pyspiel.registered_games()}
    self.assertEqual(games["kuhn_poker"].dynamics,
                     pyspiel.GameType.Dynamics.SEQUENTIAL)
    self.assertEqual(games["kuhn_poker"].chance_mode,
                     pyspiel.GameType.ChanceMode.EXPLICIT_STOCHASTIC)
    self.assertEqual(games["kuhn_poker"].information,
                     pyspiel.GameType.Information.IMPERFECT_INFORMATION)
    self.assertEqual(games["kuhn_poker"].utility,
                     pyspiel.GameType.Utility.ZERO_SUM)
    self.assertEqual(games["kuhn_poker"].min_num_players, 2)

  def test_create_game(self):
    game = pyspiel.load_game("kuhn_poker")
    game_info = game.get_type()
    self.assertEqual(game_info.information,
                     pyspiel.GameType.Information.IMPERFECT_INFORMATION)
    self.assertEqual(game.num_players(), 2)

  def test_play_kuhn_poker(self):
    game = pyspiel.load_game("kuhn_poker")
    state = game.new_initial_state()
    self.assertEqual(state.is_chance_node(), True)
    self.assertEqual(state.chance_outcomes(), [(0, 1 / 3), (1, 1 / 3),
                                               (2, 1 / 3)])
    state.apply_action(1)
    self.assertEqual(state.is_chance_node(), True)
    self.assertEqual(state.chance_outcomes(), [(0, 0.5), (2, 0.5)])
    state.apply_action(2)
    self.assertEqual(state.is_chance_node(), False)
    self.assertEqual(state.legal_actions(), [0, 1])
    sampler = pyspiel.UniformProbabilitySampler(0., 1.)
    clone = state.resample_from_infostate(1, sampler)
    self.assertEqual(
        clone.information_state_string(1), state.information_state_string(1))

  def test_othello(self):
    game = pyspiel.load_game("othello")
    state = game.new_initial_state()
    self.assertFalse(state.is_chance_node())
    self.assertFalse(state.is_terminal())
    self.assertEqual(state.legal_actions(), [19, 26, 37, 44])

  def test_tic_tac_toe(self):
    game = pyspiel.load_game("tic_tac_toe")
    state = game.new_initial_state()
    self.assertFalse(state.is_chance_node())
    self.assertFalse(state.is_terminal())
    self.assertEqual(state.legal_actions(), [0, 1, 2, 3, 4, 5, 6, 7, 8])

  def test_game_parameter_representation(self):
    param = pyspiel.GameParameter(True)
    self.assertEqual(repr(param), "GameParameter(bool_value=True)")
    param = pyspiel.GameParameter(False)
    self.assertEqual(repr(param), "GameParameter(bool_value=False)")
    param = pyspiel.GameParameter("one")
    self.assertEqual(repr(param), "GameParameter(string_value='one')")
    param = pyspiel.GameParameter(1)
    self.assertEqual(repr(param), "GameParameter(int_value=1)")
    param = pyspiel.GameParameter(1.0)
    self.assertEqual(repr(param), "GameParameter(double_value=1)")
    param = pyspiel.GameParameter(1.2)
    self.assertEqual(repr(param), "GameParameter(double_value=1.2)")

  def test_game_parameter_equality(self):
    param1 = pyspiel.GameParameter("one")
    param1_again = pyspiel.GameParameter("one")
    param2 = pyspiel.GameParameter("two")
    self.assertEqual(param1, param1_again)
    self.assertNotEqual(param1, param2)

  def test_game_parameters_from_string_empty(self):
    self.assertEqual(pyspiel.game_parameters_from_string(""), {})

  def test_game_parameters_from_string_simple(self):
    self.assertEqual(pyspiel.game_parameters_from_string("foo"),
                     {"name": pyspiel.GameParameter("foo")})

  def test_game_parameters_from_string_with_options(self):
    self.assertEqual(pyspiel.game_parameters_from_string("foo(x=2,y=true)"),
                     {"name": pyspiel.GameParameter("foo"),
                      "x": pyspiel.GameParameter(2),
                      "y": pyspiel.GameParameter(True)})

  def test_game_type(self):
    game_type = pyspiel.GameType(
        "matrix_mp", "Matching Pennies", pyspiel.GameType.Dynamics.SIMULTANEOUS,
        pyspiel.GameType.ChanceMode.DETERMINISTIC,
        pyspiel.GameType.Information.PERFECT_INFORMATION,
        pyspiel.GameType.Utility.ZERO_SUM,
        pyspiel.GameType.RewardModel.TERMINAL, 2, 2, True, True, False, False,
        dict())
    self.assertEqual(game_type.chance_mode,
                     pyspiel.GameType.ChanceMode.DETERMINISTIC)

  def test_error_handling(self):
    with six.assertRaisesRegex(self, RuntimeError,
                               "Unknown game 'invalid_game_name'"):
      unused_game = pyspiel.load_game("invalid_game_name")

  def test_can_create_cpp_tabular_policy(self):
    for game_name in ["kuhn_poker", "leduc_poker", "liars_dice"]:
      game = pyspiel.load_game(game_name)

      # We just test that we can create a tabular policy.
      policy.python_policy_to_pyspiel_policy(policy.TabularPolicy(game))

  def test_simultaneous_game_history(self):
    game = pyspiel.load_game("coop_box_pushing")
    state = game.new_initial_state()
    state.apply_action(0)
    state2 = game.new_initial_state()
    state2.apply_actions([0] * game.num_players())
    self.assertEqual(state.history(), state2.history())

  def test_record_batched_trajectories(self):
    for game_name in ["kuhn_poker", "leduc_poker", "liars_dice"]:
      game = pyspiel.load_game(game_name)
      python_policy = policy.TabularPolicy(game)
      tabular_policy = policy.python_policy_to_pyspiel_policy(python_policy)
      policies = [tabular_policy] * 2

      # We test that we can create a batch of trajectories.
      seed = 0
      batch_size = 128
      include_full_observations = False
      pyspiel.record_batched_trajectories(game, policies,
                                          python_policy.state_lookup,
                                          batch_size, include_full_observations,
                                          seed, -1)


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Reinforcement Learning (RL) Environment for Open Spiel.

This module wraps Open Spiel Python interface providing an RL-friendly API. It
covers both turn-based and simultaneous move games. Interactions between agents
and the underlying game occur mostly through the `reset` and `step` methods,
which return a `TimeStep` structure (see its docstrings for more info).

The following example illustrates the interaction dynamics. Consider a 2-player
Kuhn Poker (turn-based game). Agents have access to the `observations` (a dict)
field from `TimeSpec`, containing the following members:
 * `info_state`: list containing the game information state for each player. The
   size of the list always correspond to the number of players. E.g.:
   [[0, 1, 0, 0, 0, 0], [1, 0, 0, 0, 0, 0]].
 * `legal_actions`: list containing legal action ID lists (one for each player).
   E.g.: [[0, 1], [0]], which corresponds to actions 0 and 1 being valid for
   player 0 (the 1st player) and action 0 being valid for player 1 (2nd player).
 * `current_player`: zero-based integer representing the player to make a move.

At each `step` call, the environment expects a singleton list with the action
(as it's a turn-based game), e.g.: [1]. This (zero-based) action must correspond
to the player specified at `current_player`. The game (which is at decision
node) will process the action and take as many steps necessary to cover chance
nodes, halting at a new decision or final node. Finally, a new `TimeStep`is
returned to the agent.

Simultaneous-move games follow analogous dynamics. The only differences is the
environment expects a list of actions, one per player. Note the `current_player`
field is "irrelevant" here, admitting a constant value defined in spiel.h, which
defaults to -2 (module level constant `SIMULTANEOUS_PLAYER_ID`).

See open_spiel/python/examples/rl_example.py for example usages.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections

import enum
from absl import logging
import numpy as np

import pyspiel

SIMULTANEOUS_PLAYER_ID = pyspiel.PlayerId.SIMULTANEOUS


class TimeStep(
    collections.namedtuple(
        "TimeStep", ["observations", "rewards", "discounts", "step_type"])):
  """Returned with every call to `step` and `reset`.

  A `TimeStep` contains the data emitted by a game at each step of interaction.
  A `TimeStep` holds an `observation` (list of dicts, one per player),
  associated lists of `rewards`, `discounts` and a `step_type`.

  The first `TimeStep` in a sequence will have `StepType.FIRST`. The final
  `TimeStep` will have `StepType.LAST`. All other `TimeStep`s in a sequence will
  have `StepType.MID.

  Attributes:
    observations: a list of dicts containing observations per player.
    rewards: A list of scalars (one per player), or `None` if `step_type` is
      `StepType.FIRST`, i.e. at the start of a sequence.
    discounts: A list of discount values in the range `[0, 1]` (one per player),
      or `None` if `step_type` is `StepType.FIRST`.
    step_type: A `StepType` enum value.
  """
  __slots__ = ()

  def first(self):
    return self.step_type == StepType.FIRST

  def mid(self):
    return self.step_type == StepType.MID

  def last(self):
    return self.step_type == StepType.LAST

  def is_simultaneous_move(self):
    return self.observations["current_player"] == SIMULTANEOUS_PLAYER_ID

  def current_player(self):
    return self.observations["current_player"]


class StepType(enum.Enum):
  """Defines the status of a `TimeStep` within a sequence."""

  FIRST = 0  # Denotes the first `TimeStep` in a sequence.
  MID = 1  # Denotes any `TimeStep` in a sequence that is not FIRST or LAST.
  LAST = 2  # Denotes the last `TimeStep` in a sequence.

  def first(self):
    return self is StepType.FIRST

  def mid(self):
    return self is StepType.MID

  def last(self):
    return self is StepType.LAST


# Global pyspiel members
def registered_games():
  return pyspiel.registered_games()


class ChanceEventSampler(object):
  """Default sampler for external chance events."""

  def __init__(self, seed=None):
    self.seed(seed)

  def seed(self, seed=None):
    self._rng = np.random.RandomState(seed)

  def __call__(self, state):
    """Sample a chance event in the given state."""
    actions, probs = zip(*state.chance_outcomes())
    return self._rng.choice(actions, p=probs)


class ObservationType(enum.Enum):
  """Defines what kind of observation to use."""
  OBSERVATION = 0  # Use observation_tensor
  INFORMATION_STATE = 1  # Use information_state_tensor


class Environment(object):
  """Open Spiel reinforcement learning environment class."""

  def __init__(self,
               game,
               discount=1.0,
               chance_event_sampler=None,
               observation_type=None,
               include_full_state=False,
               **kwargs):
    """Constructor.

    Args:
      game: [string, pyspiel.Game] Open Spiel game name or game instance.
      discount: float, discount used in non-initial steps. Defaults to 1.0.
      chance_event_sampler: optional object with `sample_external_events` method
        to sample chance events.
      observation_type: what kind of observation to use. If not specified, will
        default to INFORMATION_STATE unless the game doesn't provide it.
      include_full_state: whether or not to include the full serialized
        OpenSpiel state in the observations (sometimes useful for debugging).
      **kwargs: dict, additional settings passed to the Open Spiel game.
    """
    self._chance_event_sampler = chance_event_sampler or ChanceEventSampler()
    self._include_full_state = include_full_state

    if isinstance(game, str):
      if kwargs:
        game_settings = {
            key: pyspiel.GameParameter(val) for (key, val) in kwargs.items()
        }
        logging.info("Using game settings: %s", game_settings)
        self._game = pyspiel.load_game(game, game_settings)
      else:
        logging.info("Using game string: %s", game)
        self._game = pyspiel.load_game(game)
    else:  # pyspiel.Game or API-compatible object.
      logging.info("Using game instance: %s", game.get_type().short_name)
      self._game = game

    self._num_players = self._game.num_players()
    self._state = None
    self._should_reset = True

    # Discount returned at non-initial steps.
    self._discounts = [discount] * self._num_players

    # Determine what observation type to use.
    if observation_type is None:
      if self._game.get_type().provides_information_state_tensor:
        observation_type = ObservationType.INFORMATION_STATE
      else:
        observation_type = ObservationType.OBSERVATION

    # Check the requested observation type is supported.
    if observation_type == ObservationType.OBSERVATION:
      if not self._game.get_type().provides_observation_tensor:
        raise ValueError("observation_tensor not supported by " + game)
    elif observation_type == ObservationType.INFORMATION_STATE:
      if not self._game.get_type().provides_information_state_tensor:
        raise ValueError("information_state_tensor not supported by " + game)
    self._use_observation = (observation_type == ObservationType.OBSERVATION)

  def seed(self, seed=None):
    self._chance_event_sampler.seed(seed)

  def get_time_step(self):
    """Returns a `TimeStep` without updating the environment.

    Returns:
      A `TimeStep` namedtuple containing:
        observation: list of dicts containing one observations per player, each
          corresponding to `observation_spec()`.
        reward: list of rewards at this timestep, or None if step_type is
          `StepType.FIRST`.
        discount: list of discounts in the range [0, 1], or None if step_type is
          `StepType.FIRST`.
        step_type: A `StepType` value.
    """
    observations = {
        "info_state": [],
        "legal_actions": [],
        "current_player": [],
        "serialized_state": []
    }
    rewards = []
    step_type = StepType.LAST if self._state.is_terminal() else StepType.MID
    self._should_reset = step_type == StepType.LAST

    cur_rewards = self._state.rewards()
    for player_id in range(self.num_players):
      rewards.append(cur_rewards[player_id])
      observations["info_state"].append(
          self._state.observation_tensor(player_id) if self._use_observation
          else self._state.information_state_tensor(player_id))

      observations["legal_actions"].append(self._state.legal_actions(player_id))
    observations["current_player"] = self._state.current_player()
    discounts = self._discounts
    if step_type == StepType.LAST:
      # When the game is in a terminal state set the discount to 0.
      discounts = [0. for _ in discounts]

    if self._include_full_state:
      observations["serialized_state"] = pyspiel.serialize_game_and_state(
          self._game, self._state)

    return TimeStep(
        observations=observations,
        rewards=rewards,
        discounts=discounts,
        step_type=step_type)

  def step(self, actions):
    """Updates the environment according to `actions` and returns a `TimeStep`.

    If the environment returned a `TimeStep` with `StepType.LAST` at the
    previous step, this call to `step` will start a new sequence and `actions`
    will be ignored.

    This method will also start a new sequence if called after the environment
    has been constructed and `reset` has not been called. Again, in this case
    `actions` will be ignored.

    Args:
      actions: a list containing one action per player, following specifications
        defined in `action_spec()`.

    Returns:
      A `TimeStep` namedtuple containing:
        observation: list of dicts containing one observations per player, each
          corresponding to `observation_spec()`.
        reward: list of rewards at this timestep, or None if step_type is
          `StepType.FIRST`.
        discount: list of discounts in the range [0, 1], or None if step_type is
          `StepType.FIRST`.
        step_type: A `StepType` value.
    """
    assert len(actions) == self.num_actions_per_step, (
        "Invalid number of actions! Expected {}".format(
            self.num_actions_per_step))
    if self._should_reset:
      return self.reset()

    if self.is_turn_based:
      self._state.apply_action(actions[0])
    else:
      self._state.apply_actions(actions)
    self._sample_external_events()

    return self.get_time_step()

  def reset(self):
    """Starts a new sequence and returns the first `TimeStep` of this sequence.

    Returns:
      A `TimeStep` namedtuple containing:
        observations: list of dicts containing one observations per player, each
          corresponding to `observation_spec()`.
        rewards: list of rewards at this timestep, or None if step_type is
          `StepType.FIRST`.
        discounts: list of discounts in the range [0, 1], or None if step_type
          is `StepType.FIRST`.
        step_type: A `StepType` value.
    """
    self._should_reset = False
    self._state = self._game.new_initial_state()
    self._sample_external_events()

    observations = {
        "info_state": [],
        "legal_actions": [],
        "current_player": [],
        "serialized_state": []
    }
    for player_id in range(self.num_players):
      observations["info_state"].append(
          self._state.observation_tensor(player_id) if self._use_observation
          else self._state.information_state_tensor(player_id))
      observations["legal_actions"].append(self._state.legal_actions(player_id))
    observations["current_player"] = self._state.current_player()

    if self._include_full_state:
      observations["serialized_state"] = pyspiel.serialize_game_and_state(
          self._game, self._state)

    return TimeStep(
        observations=observations,
        rewards=None,
        discounts=None,
        step_type=StepType.FIRST)

  def _sample_external_events(self):
    """Sample chance events until we get to a decision node."""
    while self._state.is_chance_node():
      outcome = self._chance_event_sampler(self._state)
      self._state.apply_action(outcome)

  def observation_spec(self):
    """Defines the observation per player provided by the environment.

    Each dict member will contain its expected structure and shape. E.g.: for
    Kuhn Poker {"info_state": (6,), "legal_actions": (2,), "current_player": (),
                "serialized_state": ()}

    Returns:
      A specification dict describing the observation fields and shapes.
    """
    return dict(
        info_state=tuple([
            self._game.observation_tensor_size() if self._use_observation else
            self._game.information_state_tensor_size()
        ]),
        legal_actions=(self._game.num_distinct_actions(),),
        current_player=(),
        serialized_state=(),
    )

  def action_spec(self):
    """Defines per player action specifications.

    Specifications include action boundaries and their data type.
    E.g.: for Kuhn Poker {"num_actions": 2, "min": 0, "max":1, "dtype": int}

    Returns:
      A specification dict containing per player action properties.
    """
    return dict(
        num_actions=self._game.num_distinct_actions(),
        min=0,
        max=self._game.num_distinct_actions() - 1,
        dtype=int,
    )

  # Game properties
  @property
  def name(self):
    return self._game.get_type().short_name

  @property
  def num_players(self):
    return self._game.num_players()

  @property
  def num_actions_per_step(self):
    return 1 if self.is_turn_based else self.num_players

  # New RL calls for more advanced use cases (e.g. search + RL).
  @property
  def is_turn_based(self):
    return self._game.get_type(
    ).dynamics == pyspiel.GameType.Dynamics.SEQUENTIAL

  @property
  def max_game_length(self):
    return self._game.max_game_length()

  @property
  def is_chance_node(self):
    return self._state.is_chance_node()

  @property
  def game(self):
    return self._game

  def set_state(self, new_state):
    """Updates the game state."""
    assert new_state.get_game() == self.game, (
        "State must have been created by the same game.")
    self._state = new_state

  @property
  def get_state(self):
    return self._state



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Tabular Q-Learning on Lewis Signaling Game."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import copy
from absl import app
from absl import flags
import numpy as np
from six.moves import range

from open_spiel.python import rl_environment
from open_spiel.python import rl_tools
from open_spiel.python.algorithms import tabular_qlearner

FLAGS = flags.FLAGS

# Env parameters
flags.DEFINE_integer("num_states", 3, "Number of states and actions")
flags.DEFINE_integer("num_messages", 3, "Number of messages")
flags.DEFINE_string("payoffs", "1, 0, 0, 0, 1, 0, 0, 0, 1",
                    "Payoffs to use ('random' for random [0, 1) payoffs)")

# Alg parameters
flags.DEFINE_bool("centralized", False, "Set to use centralized learning")
flags.DEFINE_integer("num_episodes", 2000, "Number of train episodes")
flags.DEFINE_float("step_size", 0.1, "Step size for updates")
flags.DEFINE_float("eps_init", 1.0, "Initial value of epsilon")
flags.DEFINE_float("eps_final", 0.0, "Final value of epsilon")
flags.DEFINE_integer("eps_decay_steps", 1900,
                     "Number of episodes to decay epsilon")

# Misc paramters
flags.DEFINE_integer("num_runs", 100, "Number of repetitions")
flags.DEFINE_integer("log_interval", 10,
                     "Number of episodes between each logging")
flags.DEFINE_bool("plot", False, "Set to plot the graphs")
flags.DEFINE_bool("compare", False,
                  "Set to compare centralized vs decentralized")


def run_experiment(num_players, env, payoffs, centralized):
  """Run the experiments."""
  num_states = FLAGS.num_states
  num_messages = FLAGS.num_messages
  num_actions = env.action_spec()["num_actions"]

  # Results to store
  num_runs = FLAGS.num_runs
  training_episodes = FLAGS.num_episodes
  log_interval = FLAGS.log_interval
  rewards = np.zeros((num_runs, training_episodes // log_interval))
  opts = np.zeros((num_runs, training_episodes // log_interval))
  converge_point = np.zeros((num_states, num_states))
  percent_opt = 0

  # Repeat the experiment num_runs times
  for i in range(num_runs):
    eps_schedule = rl_tools.LinearSchedule(
        FLAGS.eps_init, FLAGS.eps_final, FLAGS.eps_decay_steps *
        2)  # *2 since there are 2 agent steps per episode

    agents = [
        # pylint: disable=g-complex-comprehension
        tabular_qlearner.QLearner(
            player_id=idx,
            num_actions=num_actions,
            step_size=FLAGS.step_size,
            epsilon_schedule=eps_schedule,
            centralized=centralized) for idx in range(num_players)
    ]

    # 1. Train the agents
    for cur_episode in range(training_episodes):
      time_step = env.reset()
      # Find cur_state for logging. See lewis_signaling.cc for info_state
      # details.
      cur_state = time_step.observations["info_state"][0][3:].index(1)
      while not time_step.last():
        player_id = time_step.observations["current_player"]
        agent_output = agents[player_id].step(time_step)
        time_step = env.step([agent_output.action])

      # Episode is over, step all agents with final info state.
      for agent in agents:
        agent.step(time_step)

      # Store rewards
      reward = time_step.rewards[0]
      max_reward = payoffs[cur_state].max()
      cur_idx = (i, cur_episode // log_interval)
      rewards[cur_idx] += reward / log_interval
      opts[cur_idx] += np.isclose(reward, max_reward) / log_interval

    base_info_state0 = [1.0, 0.0, 0.0] + [0.0] * num_states
    base_info_state1 = [0.0, 1.0, 0.0] + [0.0] * num_states
    if centralized:
      base_info_state0 = [base_info_state0, base_info_state0.copy()]
      base_info_state1 = [base_info_state1, base_info_state1.copy()]

    for s in range(num_states):
      info_state0 = copy.deepcopy(base_info_state0)
      if centralized:
        info_state0[0][3 + s] = 1.0
      else:
        info_state0[3 + s] = 1.0
      # pylint: disable=protected-access
      m, _ = agents[0]._epsilon_greedy(
          str(info_state0), np.arange(num_messages), 0)
      info_state1 = copy.deepcopy(base_info_state1)
      if centralized:
        info_state1[0][3 + s] = 1.0
        info_state1[1][3 + m] = 1.0
      else:
        info_state1[3 + m] = 1.0
      a, _ = agents[1]._epsilon_greedy(
          str(info_state1), np.arange(num_states), 0)
      converge_point[s, a] += 1
      best_act = payoffs[s].argmax()
      percent_opt += int(a == best_act) / num_runs / num_states
  return rewards, opts, converge_point, percent_opt


def main(_):
  game = "lewis_signaling"
  num_players = 2

  num_states = FLAGS.num_states
  num_messages = FLAGS.num_messages
  if FLAGS.payoffs == "random":
    payoffs = np.random.random((num_states, num_states))
    payoffs_str = ",".join([str(x) for x in payoffs.flatten()])
  elif FLAGS.payoffs == "climbing":
    # This is a particular payoff matrix that is hard for decentralized
    # algorithms. Introduced in C. Claus and C. Boutilier, "The dynamics of
    # reinforcement learning in cooperative multiagent systems", 1998, for
    # simultaneous action games, but it is difficult even in the case of
    # signaling games.
    payoffs = np.array([[11, -30, 0], [-30, 7, 6], [0, 0, 5]]) / 30
    payoffs_str = ",".join([str(x) for x in payoffs.flatten()])
  else:
    payoffs_str = FLAGS.payoffs
    try:
      payoffs_list = [float(x) for x in payoffs_str.split(",")]
      payoffs = np.array(payoffs_list).reshape((num_states, num_states))
    except ValueError:
      raise ValueError(
          "There should be {} (states * actions) elements in payoff. Found {} elements"
          .format(num_states * num_states, len(payoffs_list)))

  env_configs = {
      "num_states": num_states,
      "num_messages": num_messages,
      "payoffs": payoffs_str
  }

  env = rl_environment.Environment(game, **env_configs)

  if FLAGS.compare:
    rewards_list = []
    opts_list = []
    converge_point_list = []
    percent_opt_list = []
    for centralized in [True, False]:
      rewards, opts, converge_point, percent_opt = run_experiment(
          num_players, env, payoffs, centralized)
      rewards_list += [rewards]
      opts_list += [opts]
      converge_point_list += [converge_point]
      percent_opt_list += [percent_opt]
  else:
    rewards, opts, converge_point, percent_opt = run_experiment(
        num_players, env, payoffs, FLAGS.centralized)
    rewards_list = [rewards]
    opts_list = [opts]
    converge_point_list = [converge_point]
    percent_opt_list = [percent_opt]

  if FLAGS.plot:
    # pylint: disable=g-import-not-at-top
    import matplotlib as mpl
    import matplotlib.pyplot as plt
    import scipy.stats as stats

    params = {
        "font.size": 12,
        "axes.labelsize": 12,
        "xtick.labelsize": 11,
        "ytick.labelsize": 11,
    }
    mpl.rcParams.update(params)

    def init_fig():
      fig, ax = plt.subplots(1, 1)
      ax.spines["top"].set_visible(False)
      ax.spines["right"].set_visible(False)
      return fig, ax

    def plot_scalars(scalars,
                     repetition_axis=0,
                     scalar_labels=None,
                     title=None,
                     ax_labels=None):
      """Plots scalar on ax by filling 1 standard error.

      Args:
          scalars: List of scalars to plot (mean taken over repetition
            axis)
          repetition_axis: Axis to take the mean over
          scalar_labels: Labels for the scalars (for legend)
          title: Figure title
          ax_labels: Labels for x and y axis (list of 2 strings)
      """
      if not all([len(s.shape) == 2 for s in scalars]):
        raise ValueError("Only 2D arrays supported for plotting")

      if scalar_labels is None:
        scalar_labels = [None] * len(scalars)

      if len(scalars) != len(scalar_labels):
        raise ValueError(
            "Wrong number of scalar labels, expected {} but received {}".format(
                len(scalars), len(scalar_labels)))

      _, plot_axis = init_fig()
      for i, scalar in enumerate(scalars):
        xs = np.arange(scalar.shape[1 - repetition_axis]) * FLAGS.log_interval
        mean = scalar.mean(axis=repetition_axis)
        sem = stats.sem(scalar, axis=repetition_axis)
        plot_axis.plot(xs, mean, label=scalar_labels[i])
        plot_axis.fill_between(xs, mean - sem, mean + sem, alpha=0.5)

      if title is not None:
        plot_axis.set_title(title)
      if ax_labels is not None:
        plot_axis.set_xlabel(ax_labels[0])
        plot_axis.set_ylabel(ax_labels[1])

    def plot_confusion_matrix(cm, cmap=plt.cm.Blues, title=None):
      """Plots the confusion matrix.

      Args:
          cm (np.ndarray): Confusion matrix to plot
          cmap: Color map to be used in matplotlib's imshow
          title: Figure title

      Returns:
          Figure and axis on which the confusion matrix is plotted
      """
      fig, ax = plt.subplots()
      ax.imshow(cm, interpolation="nearest", cmap=cmap)
      ax.set_xticks([])
      ax.set_yticks([])
      ax.set_xlabel("Receiver's action", fontsize=14)
      ax.set_ylabel("Sender's state", fontsize=14)
      # Loop over data dimensions and create text annotations.
      fmt = "d"
      thresh = cm.max() / 2.
      for i in range(cm.shape[0]):
        for j in range(cm.shape[1]):
          ax.text(
              j,
              i,
              format(cm[i, j], fmt),
              ha="center",
              va="center",
              color="white" if cm[i, j] > thresh else "black")
      fig.tight_layout()
      if title is not None:
        ax.set_title(title)
      return fig, ax

    if FLAGS.compare:
      labels = ["Centralized", "Decentralized"]
    else:
      labels = ["Centralized"] if FLAGS.centralized else ["Decentralized"]
    plot_scalars(
        rewards_list,
        scalar_labels=labels,
        title="Reward graph (Tabular Q-Learning)",
        ax_labels=["Episodes", "Reward per episode"])
    plt.legend()
    plot_scalars(
        opts_list,
        scalar_labels=labels,
        title="Percentage of optimal actions (Tabular Q-Learning)",
        ax_labels=["Episodes", "% optimal actions"])
    plt.legend()

    for i, cp in enumerate(converge_point_list):
      plot_confusion_matrix(
          cp.astype(np.int),
          title="Final policy (Tabular {})".format(labels[i]))

    plt.show()

  return percent_opt_list


if __name__ == "__main__":
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""An example game written in Python.

Games are strongly encouraged to be written in C++ for efficiency and so that
they can be available from all APIs. Writing Python-only games means that they
they will be slow and only be accessible from Python algorithms.

Nonetheless, there are many cases where this is desirable (or required), hence
we provide this example.
"""

import copy
import pickle

import numpy as np

import pyspiel

_NUM_PLAYERS = 2
_NUM_ROWS = 3
_NUM_COLS = 3
_NUM_CELLS = _NUM_ROWS * _NUM_COLS


def _line_value(line):
  """Checks a possible line, returning the winning symbol if any."""
  if all(line == "x") or all(line == "o"):
    return line[0]


class TicTacToeState(object):
  """A python-only version of the Tic-Tac-Toe state.

  This class implements all the pyspiel.State API functions. Please see spiel.h
  for more thorough documentation of each function.

  Note that this class does not inherit from pyspiel.State since pickle
  serialization is not possible due to what is required on the C++ side
  (backpointers to the C++ game object, which we can't get from here).
  """

  def __init__(self, game):
    self._game = game
    self._cur_player = 0
    self._winner = None
    self._is_terminal = False
    self._history = []
    self._board = np.full((_NUM_ROWS, _NUM_COLS), ".")

  # Helper functions (not part of the OpenSpiel API).

  def _coord(self, move):
    return (move // _NUM_COLS, move % _NUM_COLS)

  def _line_exists(self):
    """Checks if a line exists, returns "x" or "o" if so, and None otherwise."""
    return (_line_value(self._board[0]) or _line_value(self._board[1]) or
            _line_value(self._board[2]) or _line_value(self._board[:, 0]) or
            _line_value(self._board[:, 1]) or _line_value(self._board[:, 2]) or
            _line_value(self._board.diagonal()) or
            _line_value(np.fliplr(self._board).diagonal()))

  # OpenSpiel (PySpiel) API functions are below. These need to be provided by
  # every game. Some not-often-used methods have been omitted.

  def current_player(self):
    return pyspiel.PlayerId.TERMINAL if self._is_terminal else self._cur_player

  def legal_actions(self, player=None):
    """Returns a list of legal actions, sorted in ascending order.

    Args:
      player: the player whose legal moves

    Returns:
      A list of legal moves, where each move is in [0, num_distinct_actions - 1]
      at non-terminal states, and empty list at terminal states.
    """

    if player is not None and player != self._cur_player:
      return []
    elif self.is_terminal():
      return []
    else:
      actions = []
      for action in range(_NUM_CELLS):
        if self._board[self._coord(action)] == ".":
          actions.append(action)
      return actions

  def legal_actions_mask(self, player=None):
    """Get a list of legal actions.

    Args:
      player: the player whose moves we want; defaults to the current player.

    Returns:
      A list of legal moves, where each move is in [0, num_distinct_actios - 1].
      Returns an empty list at terminal states, or if it is not the specified
      player's turn.
    """
    if player is not None and player != self._cur_player:
      return []
    elif self.is_terminal():
      return []
    else:
      action_mask = [0] * _NUM_CELLS
      for action in self.legal_actions():
        action_mask[action] = 1
      return action_mask

  def apply_action(self, action):
    """Applies the specified action to the state."""
    self._board[self._coord(action)] = "x" if self._cur_player == 0 else "o"
    self._history.append(action)
    if self._line_exists():
      self._is_terminal = True
      self._winner = self._cur_player
    elif len(self._history) == _NUM_CELLS:
      self._is_terminal = True
    else:
      self._cur_player = 1 - self._cur_player

  def undo_action(self, action):
    # Optional function. Not used in many places.
    self._board[self._coord(action)] = "."
    self._cur_player = 1 - self._cur_player
    self._history.pop()
    self._winner = None
    self._is_terminal = False

  def action_to_string(self, arg0, arg1=None):
    """Action -> string. Args either (player, action) or (action)."""
    player = self.current_player() if arg1 is None else arg0
    action = arg0 if arg1 is None else arg1
    row, col = self._coord(action)
    return "{}({},{})".format("x" if player == 0 else "o", row, col)

  def is_terminal(self):
    return self._is_terminal

  def returns(self):
    if self.is_terminal():
      if self._winner == 0:
        return [1.0, -1.0]
      elif self._winner == 1:
        return [-1.0, 1.0]
    return [0.0, 0.0]

  def rewards(self):
    return self.returns()

  def player_reward(self, player):
    return self.rewards()[player]

  def player_returns(self, player):
    return self.returns()[player]

  def is_chance_node(self):
    return False

  def is_simultaneous_node(self):
    return False

  def history(self):
    return self._history

  def history_str(self):
    return str(self._history)

  def child(self, action):
    cloned_state = self.clone()
    cloned_state.apply_action(action)
    return cloned_state

  def apply_actions(self, actions):
    raise NotImplementedError  # Only applies to simultaneous move games

  def num_distinct_actions(self):
    return _NUM_CELLS

  def num_players(self):
    return _NUM_PLAYERS

  def chance_outcomes(self):
    return []

  def get_game(self):
    return self._game

  def get_type(self):
    return self._game.get_type()

  def serialize(self):
    return pickle.dumps(self)

  def resample_from_infostate(self):
    return [self.clone()]

  def __str__(self):
    return "\n".join("".join(row) for row in self._board)

  def clone(self):
    return copy.deepcopy(self)


class TicTacToeGame(object):
  """A python-only version of the Tic-Tac-Toe game.

  This class implements all the pyspiel.Gae API functions. Please see spiel.h
  for more thorough documentation of each function.

  Note that this class does not inherit from pyspiel.Game since pickle
  serialization is not possible due to what is required on the C++ side
  (backpointers to the C++ game object, which we can't get from here).
  """

  def __init__(self):
    pass

  def new_initial_state(self):
    return TicTacToeState(self)

  def num_distinct_actions(self):
    return _NUM_CELLS

  def policy_tensor_shape(self):
    return (_NUM_ROWS, _NUM_COLS, 1)

  def clone(self):
    return TicTacToeGame()

  def max_chance_outcomes(self):
    return 0

  def get_parameters(self):
    return {}

  def num_players(self):
    return _NUM_PLAYERS

  def min_utility(self):
    return -1.0

  def max_utility(self):
    return 1.0

  def get_type(self):
    return pyspiel.GameType(
        short_name="python_tic_tac_toe",
        long_name="Python Tic-Tac-Toe",
        dynamics=pyspiel.GameType.Dynamics.SEQUENTIAL,
        chance_mode=pyspiel.GameType.ChanceMode.DETERMINISTIC,
        information=pyspiel.GameType.Information.PERFECT_INFORMATION,
        utility=pyspiel.GameType.Utility.ZERO_SUM,
        reward_model=pyspiel.GameType.RewardModel.TERMINAL,
        max_num_players=_NUM_PLAYERS,
        min_num_players=_NUM_PLAYERS,
        provides_information_state_string=True,
        provides_information_state_tensor=False,
        provides_observation_string=True,
        provides_observation_tensor=True,
        parameter_specification={},
    )

  def utility_sum(self):
    return 0.0

  def observation_tensor_layout(self):
    return pyspiel.TensorLayout.CHW

  def deserialize_state(self, string):
    return pickle.loads(string)

  def max_game_length(self):
    return _NUM_CELLS

  def __str__(self):
    return "python_tic_tac_toe"

  def make_py_observer(self, iig_obs_type, params):
    if params:
      raise ValueError("Params not supported")
    if iig_obs_type:
      raise ValueError("Not an imperfect information game")
    return TicTacToeObserver()


class TicTacToeObserver:
  """Observer, conforming to the PyObserver interface (see observer.py)."""

  def __init__(self):
    self._obs = np.zeros((3, 3, 3), np.float32)
    self.tensor = np.ravel(self._obs)
    self.dict = {"observation": self._obs}

  def set_from(self, state, player):
    del player
    self._obs[:] = 0
    board = state._board  # pylint: disable=protected-access
    for row in range(_NUM_ROWS):
      for col in range(_NUM_COLS):
        index = ".ox".index(board[row, col])
        self._obs[index, row, col] = 1.0

  def string_from(self, state, player):
    del player
    return str(state)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Tests for Python Tic-Tac-Toe."""

from absl.testing import absltest
import numpy as np
from open_spiel.python.games import tic_tac_toe


class TicTacToeTest(absltest.TestCase):

  def test_create(self):
    game = tic_tac_toe.TicTacToeGame()
    print(game.num_distinct_actions())
    clone = game.clone()
    print(clone.num_distinct_actions())
    state = game.new_initial_state()
    clone = state.clone()
    print(state)
    print(clone)

  def test_random_game(self):
    # This is here mostly to show the API by example.
    # More serious simulation tests are done in python/tests/game_sim_test.py.
    # Those test the conformance to the API thoroughly.
    game = tic_tac_toe.TicTacToeGame()
    state = game.new_initial_state()
    while not state.is_terminal():
      print(state)
      cur_player = state.current_player()
      legal_actions = state.legal_actions()
      action = np.random.choice(legal_actions)
      print("Player {} chooses action {}".format(cur_player, action))
      state.apply_action(action)
    print(state)
    print("Returns: {}".format(state.returns()))


if __name__ == "__main__":
  absltest.main()



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""Perturbates any policy with tabular-saved, fixed noise.

The policy's probabilities P' on each state s are computed as

P'(s) = alpha * epsilon + (1-alpha) * P(s),

with P the former policy's probabilities, and epsilon ~ Softmax(beta *
Uniform)
"""

import numpy as np
from open_spiel.python import policy as openspiel_policy


class NoisyPolicy(openspiel_policy.Policy):
  """Pyspiel Best Response with added noise.

    This policy's probabilities P' on each `player_id` state s is computed as
    P'(s) = alpha * epsilon + (1-alpha) * P(s),

    with P the former policy's probabilities, and epsilon ~ Softmax(beta *
    Uniform)
  """

  def __init__(self, policy, player_id=None, alpha=0.1,
               beta=1.0):
    """Initializes the noisy policy.

    Note that this noise only affects `player_id`.

    Args:
      policy: Any OpenSpiel `policy.Policy` object.
      player_id: The player id, the policy of whom will be made noisy. If `None`
        noise will be added to the policies for all players.
      alpha: Mixing noise factor.
      beta: Softmax inverse temperature factor.
    """
    self._policy = policy
    self.game = policy.game
    self.game_type = self.game.get_type()

    self.player_id = player_id

    self._noise_dict = {}
    self._alpha = alpha
    self._beta = beta

  def _state_key(self, state, player):
    """Returns the key to use to look up this (state, player) pair."""
    if self.game_type.provides_information_state_string:
      if player is None:
        return state.information_state_string()
      else:
        return state.information_state_string(player)
    elif self.game_type.provides_observation_string:
      if player is None:
        return state.observation_string()
      else:
        return state.observation_string(player)
    else:
      return str(state)

  def get_or_create_noise(self, state):
    info_state = self._state_key(state, state.current_player())
    if info_state not in self._noise_dict:
      action_ids = state.legal_actions()
      noise = self._beta * np.random.normal(size=len(action_ids))
      noise = np.exp(noise - noise.max())
      noise /= np.sum(noise)

      self._noise_dict[info_state] = {
          action_ids[i]: noise[i] for i in range(len(noise))
      }
    return self._noise_dict[info_state]

  def mix_probs(self, probs, noise_probs):
    return {
        i: (1 - self._alpha) * probs[i] + self._alpha * noise_probs[i]
        for i in probs
    }

  @property
  def policy(self):
    return self._policy

  def action_probabilities(self, state, player_id=None):
    """Returns the policy for a player in a state.

    Args:
      state: A `pyspiel.State` object.
      player_id: Optional, the player id for whom we want an action. Optional
        unless this is a simultabeous state at which multiple players can act.

    Returns:
      A `dict` of `{action: probability}` for the specified player in the
      supplied state.
    """

    # If self._player_id is None, or if self.player_id == current_player, add
    # noise.
    if (not self.player_id) or (state.current_player() == self.player_id):
      noise_probs = self.get_or_create_noise(state)
      probs = self._policy.action_probabilities(state, player_id)
      probs = self.mix_probs(probs, noise_probs)
      return probs

    # Send the default probabilities for all other players
    return self._policy.action_probabilities(state, player_id)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python spiel example."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
from absl import app
from absl import flags

from open_spiel.python import rl_environment
from open_spiel.python.algorithms import random_agent

FLAGS = flags.FLAGS

flags.DEFINE_string("game", "kuhn_poker", "Name of the game.")
flags.DEFINE_integer("num_players", 2, "Number of players.")
flags.DEFINE_integer("num_episodes", 2, "Number of episodes.")


def print_iteration(time_step, player_id, action=None):
  """Print TimeStep information."""
  obs = time_step.observations
  logging.info("Player: %s", player_id)
  if time_step.first():
    logging.info("Info state: %s, - - %s", obs["info_state"][player_id],
                 time_step.step_type)
  else:
    logging.info("Info state: %s, %s %s %s", obs["info_state"][player_id],
                 time_step.rewards[player_id], time_step.discounts[player_id],
                 time_step.step_type)
  if action is not None:
    logging.info("Action taken: %s", action)
  logging.info("-" * 80)


def main_loop(unused_arg):
  """RL main loop example."""
  logging.info("Registered games: %s", rl_environment.registered_games())
  logging.info("Creating game %s", FLAGS.game)

  env_configs = {"players": FLAGS.num_players} if FLAGS.num_players else {}
  env = rl_environment.Environment(FLAGS.game, **env_configs)
  num_actions = env.action_spec()["num_actions"]

  agents = [
      random_agent.RandomAgent(player_id=i, num_actions=num_actions)
      for i in range(FLAGS.num_players)
  ]

  logging.info("Env specs: %s", env.observation_spec())
  logging.info("Action specs: %s", env.action_spec())

  for cur_episode in range(FLAGS.num_episodes):
    logging.info("Starting episode %s", cur_episode)
    time_step = env.reset()
    while not time_step.last():
      pid = time_step.observations["current_player"]

      if env.is_turn_based:
        agent_output = agents[pid].step(time_step)
        action_list = [agent_output.action]
      else:
        agents_output = [agent.step(time_step) for agent in agents]
        action_list = [agent_output.action for agent_output in agents_output]

      print_iteration(time_step, pid, action_list)
      time_step = env.step(action_list)

    # Episode is over, step all agents with final state.
    for agent in agents:
      agent.step(time_step)

    # Print final state of end game.
    for pid in range(env.num_players):
      print_iteration(time_step, pid)


if __name__ == "__main__":
  app.run(main_loop)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as: python3
"""DQN as a policy.

Treating RL Oracles as policies allows us to streamline their use with tabular
policies and other policies in OpenSpiel, and freely mix populations using
different types of oracles.
"""

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import dqn
from open_spiel.python.algorithms import policy_gradient


def rl_policy_factory(rl_class):
  """Transforms an RL Agent into an OpenSpiel policy.

  Args:
    rl_class: An OpenSpiel class inheriting from 'rl_agent.AbstractAgent' such
      as policy_gradient.PolicyGradient or dqn.DQN.

  Returns:
    An RLPolicy class that wraps around an instance of rl_class to transform it
    into a policy.
  """

  class RLPolicy(policy.Policy):
    """A 'policy.Policy' wrapper around an 'rl_agent.AbstractAgent' instance."""

    def __init__(self, env, player_id, **kwargs):
      """Constructs an RL Policy.

      Args:
        env: An OpenSpiel RL Environment instance.
        player_id: The ID of the DQN policy's player.
        **kwargs: Various kwargs used to initialize rl_class.
      """
      game = env.game

      super(RLPolicy, self).__init__(game, player_id)
      self._policy = rl_class(**{"player_id": player_id, **kwargs})

      self._frozen = False
      self._rl_class = rl_class
      self._env = env
      self._obs = {
          "info_state": [None] * self.game.num_players(),
          "legal_actions": [None] * self.game.num_players()
      }

    def get_time_step(self):
      time_step = self._env.get_time_step()
      return time_step

    def action_probabilities(self, state, player_id=None):
      cur_player = state.current_player()
      legal_actions = state.legal_actions(cur_player)

      step_type = rl_environment.StepType.LAST if state.is_terminal(
      ) else rl_environment.StepType.MID

      self._obs["current_player"] = cur_player
      self._obs["info_state"][cur_player] = (
          state.information_state_tensor(cur_player))
      self._obs["legal_actions"][cur_player] = legal_actions

      # pylint: disable=protected-access
      rewards = state.rewards()
      if rewards:
        time_step = rl_environment.TimeStep(
            observations=self._obs, rewards=rewards,
            discounts=self._env._discounts, step_type=step_type)
      else:
        rewards = [0] * self._num_players
        time_step = rl_environment.TimeStep(
            observations=self._obs, rewards=rewards,
            discounts=self._env._discounts,
            step_type=rl_environment.StepType.FIRST)
      # pylint: enable=protected-access

      p = self._policy.step(time_step, is_evaluation=True).probs
      prob_dict = {action: p[action] for action in legal_actions}
      return prob_dict

    def step(self, time_step, is_evaluation=False):
      # The _frozen attribute freezes the weights of the current policy. This
      # effect is achieved by considering that we always are evaluating when the
      # current policy's weights are frozen. For more details, see the freeze()
      # method.
      is_evaluation = (is_evaluation) or (self._frozen)
      return self._policy.step(time_step, is_evaluation)

    def freeze(self):
      """This method freezes the policy's weights.

      The weight freezing effect is implemented by preventing any training to
      take place through calls to the step function. The weights are therefore
      not effectively frozen, and unconventional calls may trigger weights
      training.

      The weight-freezing effect is especially needed in PSRO, where all
      policies that aren't being trained by the oracle must be static. Freezing
      trained policies permitted us not to change how 'step' was called when
      introducing self-play (By not changing 'is_evaluation' depending on the
      current player).
      """
      self._frozen = True

    def unfreeze(self):
      self._frozen = False

    def is_frozen(self):
      return self._frozen

    def get_weights(self):
      return self._policy.get_weights()

    def copy_with_noise(self, sigma=0.0):
      copied_object = RLPolicy.__new__(RLPolicy)
      super(RLPolicy, copied_object).__init__(self.game, self.player_ids)
      setattr(copied_object, "_rl_class", self._rl_class)
      setattr(copied_object, "_obs", self._obs)
      setattr(copied_object, "_policy",
              self._policy.copy_with_noise(sigma=sigma))
      setattr(copied_object, "_env", self._env)
      copied_object.unfreeze()

      return copied_object

  return RLPolicy


# Generating policy classes for Policy Gradient and DQN
# pylint: disable=invalid-name
PGPolicy = rl_policy_factory(policy_gradient.PolicyGradient)
DQNPolicy = rl_policy_factory(dqn.DQN)
# pylint: enable=invalid-name



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Implements Deep CFR Algorithm.

See https://arxiv.org/abs/1811.00164.

The algorithm defines an `advantage` and `strategy` networks that compute
advantages used to do regret matching across information sets and to approximate
the strategy profiles of the game. To train these networks a reservoir buffer
(other data structures may be used) memory is used to accumulate samples to
train the networks.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import random
import numpy as np
import tensorflow.compat.v1 as tf

from open_spiel.python import policy
from open_spiel.python import simple_nets
import pyspiel

# Temporarily Disable TF2 behavior until we update the code.
tf.disable_v2_behavior()

AdvantageMemory = collections.namedtuple(
    "AdvantageMemory", "info_state iteration advantage action")

StrategyMemory = collections.namedtuple(
    "StrategyMemory", "info_state iteration strategy_action_probs")


# TODO(author3) Refactor into data structures lib.
class ReservoirBuffer(object):
  """Allows uniform sampling over a stream of data.

  This class supports the storage of arbitrary elements, such as observation
  tensors, integer actions, etc.

  See https://en.wikipedia.org/wiki/Reservoir_sampling for more details.
  """

  def __init__(self, reservoir_buffer_capacity):
    self._reservoir_buffer_capacity = reservoir_buffer_capacity
    self._data = []
    self._add_calls = 0

  def add(self, element):
    """Potentially adds `element` to the reservoir buffer.

    Args:
      element: data to be added to the reservoir buffer.
    """
    if len(self._data) < self._reservoir_buffer_capacity:
      self._data.append(element)
    else:
      idx = np.random.randint(0, self._add_calls + 1)
      if idx < self._reservoir_buffer_capacity:
        self._data[idx] = element
    self._add_calls += 1

  def sample(self, num_samples):
    """Returns `num_samples` uniformly sampled from the buffer.

    Args:
      num_samples: `int`, number of samples to draw.

    Returns:
      An iterable over `num_samples` random elements of the buffer.

    Raises:
      ValueError: If there are less than `num_samples` elements in the buffer
    """
    if len(self._data) < num_samples:
      raise ValueError("{} elements could not be sampled from size {}".format(
          num_samples, len(self._data)))
    return random.sample(self._data, num_samples)

  def clear(self):
    self._data = []
    self._add_calls = 0

  def __len__(self):
    return len(self._data)

  def __iter__(self):
    return iter(self._data)


class DeepCFRSolver(policy.Policy):
  """Implements a solver for the Deep CFR Algorithm.

  See https://arxiv.org/abs/1811.00164.

  Define all networks and sampling buffers/memories.  Derive losses & learning
  steps. Initialize the game state and algorithmic variables.

  Note: batch sizes default to `None` implying that training over the full
        dataset in memory is done by default.  To sample from the memories you
        may set these values to something less than the full capacity of the
        memory.
  """

  def __init__(self,
               session,
               game,
               policy_network_layers=(256, 256),
               advantage_network_layers=(128, 128),
               num_iterations: int = 100,
               num_traversals: int = 20,
               learning_rate: float = 1e-4,
               batch_size_advantage=None,
               batch_size_strategy=None,
               memory_capacity: int = int(1e6),
               policy_network_train_steps: int = 1,
               advantage_network_train_steps: int = 1,
               reinitialize_advantage_networks: bool = True):
    """Initialize the Deep CFR algorithm.

    Args:
      session: (tf.Session) TensorFlow session.
      game: Open Spiel game.
      policy_network_layers: (list[int]) Layer sizes of strategy net MLP.
      advantage_network_layers: (list[int]) Layer sizes of advantage net MLP.
      num_iterations: Number of iterations.
      num_traversals: Number of traversals per iteration.
      learning_rate: Learning rate.
      batch_size_advantage: (int or None) Batch size to sample from advantage
        memories.
      batch_size_strategy: (int or None) Batch size to sample from strategy
        memories.
      memory_capacity: Number of samples that can be stored in memory.
      policy_network_train_steps: Number of policy network training steps (per
        iteration).
      advantage_network_train_steps: Number of advantage network training steps
        (per iteration).
      reinitialize_advantage_networks: Whether to re-initialize the
        advantage network before training on each iteration.
    """
    all_players = list(range(game.num_players()))
    super(DeepCFRSolver, self).__init__(game, all_players)
    self._game = game
    if game.get_type().dynamics == pyspiel.GameType.Dynamics.SIMULTANEOUS:
      # `_traverse_game_tree` does not take into account this option.
      raise ValueError("Simulatenous games are not supported.")
    self._session = session
    self._batch_size_advantage = batch_size_advantage
    self._batch_size_strategy = batch_size_strategy
    self._policy_network_train_steps = policy_network_train_steps
    self._advantage_network_train_steps = advantage_network_train_steps
    self._num_players = game.num_players()
    self._root_node = self._game.new_initial_state()
    # TODO(author6) Allow embedding size (and network) to be specified.
    self._embedding_size = len(self._root_node.information_state_tensor(0))
    self._num_iterations = num_iterations
    self._num_traversals = num_traversals
    self._reinitialize_advantage_networks = reinitialize_advantage_networks
    self._num_actions = game.num_distinct_actions()
    self._iteration = 1

    # Create required TensorFlow placeholders to perform the Q-network updates.
    self._info_state_ph = tf.placeholder(
        shape=[None, self._embedding_size],
        dtype=tf.float32,
        name="info_state_ph")
    self._info_state_action_ph = tf.placeholder(
        shape=[None, self._embedding_size + 1],
        dtype=tf.float32,
        name="info_state_action_ph")
    self._action_probs_ph = tf.placeholder(
        shape=[None, self._num_actions],
        dtype=tf.float32,
        name="action_probs_ph")
    self._iter_ph = tf.placeholder(
        shape=[None, 1], dtype=tf.float32, name="iter_ph")
    self._advantage_ph = []
    for p in range(self._num_players):
      self._advantage_ph.append(
          tf.placeholder(
              shape=[None, self._num_actions],
              dtype=tf.float32,
              name="advantage_ph_" + str(p)))

    # Define strategy network, loss & memory.
    self._strategy_memories = ReservoirBuffer(memory_capacity)
    self._policy_network = simple_nets.MLP(self._embedding_size,
                                           list(policy_network_layers),
                                           self._num_actions)
    action_logits = self._policy_network(self._info_state_ph)
    # Illegal actions are handled in the traversal code where expected payoff
    # and sampled regret is computed from the advantage networks.
    self._action_probs = tf.nn.softmax(action_logits)
    self._loss_policy = tf.reduce_mean(
        tf.losses.mean_squared_error(
            labels=tf.math.sqrt(self._iter_ph) * self._action_probs_ph,
            predictions=tf.math.sqrt(self._iter_ph) * self._action_probs))
    self._optimizer_policy = tf.train.AdamOptimizer(learning_rate=learning_rate)
    self._learn_step_policy = self._optimizer_policy.minimize(self._loss_policy)

    # Define advantage network, loss & memory. (One per player)
    self._advantage_memories = [
        ReservoirBuffer(memory_capacity) for _ in range(self._num_players)
    ]
    self._advantage_networks = [
        simple_nets.MLP(self._embedding_size, list(advantage_network_layers),
                        self._num_actions) for _ in range(self._num_players)
    ]
    self._advantage_outputs = [
        self._advantage_networks[i](self._info_state_ph)
        for i in range(self._num_players)
    ]
    self._loss_advantages = []
    self._optimizer_advantages = []
    self._learn_step_advantages = []
    for p in range(self._num_players):
      self._loss_advantages.append(
          tf.reduce_mean(
              tf.losses.mean_squared_error(
                  labels=tf.math.sqrt(self._iter_ph) * self._advantage_ph[p],
                  predictions=tf.math.sqrt(self._iter_ph) *
                  self._advantage_outputs[p])))
      self._optimizer_advantages.append(
          tf.train.AdamOptimizer(learning_rate=learning_rate))
      self._learn_step_advantages.append(self._optimizer_advantages[p].minimize(
          self._loss_advantages[p]))

  @property
  def advantage_buffers(self):
    return self._advantage_memories

  @property
  def strategy_buffer(self):
    return self._strategy_memories

  def clear_advantage_buffers(self):
    for p in range(self._num_players):
      self._advantage_memories[p].clear()

  def reinitialize_advantage_networks(self):
    for p in range(self._num_players):
      self.reinitialize_advantage_network(p)

  def reinitialize_advantage_network(self, player):
    self._session.run(
        tf.group(*[
            var.initializer
            for var in self._advantage_networks[player].variables
        ]))

  def solve(self):
    """Solution logic for Deep CFR."""
    advantage_losses = collections.defaultdict(list)
    for _ in range(self._num_iterations):
      for p in range(self._num_players):
        for _ in range(self._num_traversals):
          self._traverse_game_tree(self._root_node, p)
        if self._reinitialize_advantage_networks:
          # Re-initialize advantage network for player and train from scratch.
          self.reinitialize_advantage_network(p)
        advantage_losses[p].append(self._learn_advantage_network(p))
      self._iteration += 1
    # Train policy network.
    policy_loss = self._learn_strategy_network()
    return self._policy_network, advantage_losses, policy_loss

  def _traverse_game_tree(self, state, player):
    """Performs a traversal of the game tree.

    Over a traversal the advantage and strategy memories are populated with
    computed advantage values and matched regrets respectively.
    Args:
      state: Current OpenSpiel game state.
      player: (int) Player index for this traversal.
    Returns:
      Recursively returns expected payoffs for each action.
    """
    expected_payoff = collections.defaultdict(float)
    if state.is_terminal():
      # Terminal state get returns.
      return state.returns()[player]
    elif state.is_chance_node():
      # If this is a chance node, sample an action
      action = np.random.choice([i[0] for i in state.chance_outcomes()])
      return self._traverse_game_tree(state.child(action), player)
    elif state.current_player() == player:
      sampled_regret = collections.defaultdict(float)
      # Update the policy over the info set & actions via regret matching.
      _, strategy = self._sample_action_from_advantage(state, player)
      for action in state.legal_actions():
        expected_payoff[action] = self._traverse_game_tree(
            state.child(action), player)
      cfv = 0
      for a_ in state.legal_actions():
        cfv += strategy[a_] * expected_payoff[a_]
      for action in state.legal_actions():
        sampled_regret[action] = expected_payoff[action]
        sampled_regret[action] -= cfv
      sampled_regret_arr = [0] * self._num_actions
      for action in sampled_regret:
        sampled_regret_arr[action] = sampled_regret[action]
      self._advantage_memories[player].add(
          AdvantageMemory(state.information_state_tensor(), self._iteration,
                          sampled_regret_arr, action))
      return cfv
    else:
      other_player = state.current_player()
      _, strategy = self._sample_action_from_advantage(state, other_player)
      # Recompute distribution dor numerical errors.
      probs = np.array(strategy)
      probs /= probs.sum()
      sampled_action = np.random.choice(range(self._num_actions), p=probs)
      self._strategy_memories.add(
          StrategyMemory(
              state.information_state_tensor(other_player), self._iteration,
              strategy))
      return self._traverse_game_tree(state.child(sampled_action), player)

  def _sample_action_from_advantage(self, state, player):
    """Returns an info state policy by applying regret-matching.

    Args:
      state: Current OpenSpiel game state.
      player: (int) Player index over which to compute regrets.
    Returns:
      1. (list) Advantage values for info state actions indexed by action.
      2. (list) Matched regrets, prob for actions indexed by action.
    """
    info_state = state.information_state_tensor(player)
    legal_actions = state.legal_actions(player)
    advantages_full = self._session.run(
        self._advantage_outputs[player],
        feed_dict={self._info_state_ph: np.expand_dims(info_state, axis=0)})[0]
    advantages = [max(0., advantage) for advantage in advantages_full]
    cumulative_regret = np.sum([advantages[action] for action in legal_actions])
    matched_regrets = np.array([0.] * self._num_actions)

    if cumulative_regret > 0.:
      for action in legal_actions:
        matched_regrets[action] = advantages[action] / cumulative_regret
    else:
      matched_regrets[max(legal_actions, key=lambda a: advantages_full[a])] = 1

    return advantages, matched_regrets

  def action_probabilities(self, state):
    """Returns action probabilities dict for a single batch."""
    cur_player = state.current_player()
    legal_actions = state.legal_actions(cur_player)
    info_state_vector = np.array(state.information_state_tensor())
    if len(info_state_vector.shape) == 1:
      info_state_vector = np.expand_dims(info_state_vector, axis=0)
    probs = self._session.run(
        self._action_probs, feed_dict={self._info_state_ph: info_state_vector})
    return {action: probs[0][action] for action in legal_actions}

  def _learn_advantage_network(self, player):
    """Compute the loss on sampled transitions and perform a Q-network update.

    If there are not enough elements in the buffer, no loss is computed and
    `None` is returned instead.

    Args:
      player: (int) player index.
    Returns:
      The average loss over the advantage network.
    """
    for _ in range(self._advantage_network_train_steps):
      if self._batch_size_advantage:
        if self._batch_size_advantage > len(self._advantage_memories[player]):
          ## Skip if there aren't enough samples
          return None
        samples = self._advantage_memories[player].sample(
            self._batch_size_advantage)
      else:
        samples = self._advantage_memories[player]
      info_states = []
      advantages = []
      iterations = []
      for s in samples:
        info_states.append(s.info_state)
        advantages.append(s.advantage)
        iterations.append([s.iteration])
      # Ensure some samples have been gathered.
      if not info_states:
        return None

      loss_advantages, _ = self._session.run(
          [self._loss_advantages[player], self._learn_step_advantages[player]],
          feed_dict={
              self._info_state_ph: np.array(info_states),
              self._advantage_ph[player]: np.array(advantages),
              self._iter_ph: np.array(iterations),
          })
    return loss_advantages

  def _learn_strategy_network(self):
    """Compute the loss over the strategy network.

    Returns:
      The average loss obtained on this batch of transitions or `None`.
    """
    for _ in range(self._policy_network_train_steps):
      if self._batch_size_strategy:
        if self._batch_size_strategy > len(self._strategy_memories):
          ## Skip if there aren't enough samples
          return None
        samples = self._strategy_memories.sample(self._batch_size_strategy)
      else:
        samples = self._strategy_memories
      info_states = []
      action_probs = []
      iterations = []
      for s in samples:
        info_states.append(s.info_state)
        action_probs.append(s.strategy_action_probs)
        iterations.append([s.iteration])

      loss_strategy, _ = self._session.run(
          [self._loss_policy, self._learn_step_policy],
          feed_dict={
              self._info_state_ph: np.array(info_states),
              self._action_probs_ph: np.array(np.squeeze(action_probs)),
              self._iter_ph: np.array(iterations),
          })
    return loss_strategy



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""NFSP agents trained on Leduc Poker."""

from absl import app
from absl import flags
from absl import logging
import tensorflow.compat.v1 as tf

from open_spiel.python import policy
from open_spiel.python import rl_environment
from open_spiel.python.algorithms import exploitability
from open_spiel.python.algorithms import nfsp

FLAGS = flags.FLAGS

flags.DEFINE_string("game_name", "leduc_poker",
                    "Name of the game.")
flags.DEFINE_integer("num_players", 2,
                     "Number of players.")
flags.DEFINE_integer("num_train_episodes", int(20e6),
                     "Number of training episodes.")
flags.DEFINE_integer("eval_every", 10000,
                     "Episode frequency at which the agents are evaluated.")
flags.DEFINE_list("hidden_layers_sizes", [
    128,
], "Number of hidden units in the avg-net and Q-net.")
flags.DEFINE_integer("replay_buffer_capacity", int(2e5),
                     "Size of the replay buffer.")
flags.DEFINE_integer("reservoir_buffer_capacity", int(2e6),
                     "Size of the reservoir buffer.")
flags.DEFINE_integer("min_buffer_size_to_learn", 1000,
                     "Number of samples in buffer before learning begins.")
flags.DEFINE_float("anticipatory_param", 0.1,
                   "Prob of using the rl best response as episode policy.")
flags.DEFINE_integer("batch_size", 128,
                     "Number of transitions to sample at each learning step.")
flags.DEFINE_integer("learn_every", 64,
                     "Number of steps between learning updates.")
flags.DEFINE_float("rl_learning_rate", 0.01,
                   "Learning rate for inner rl agent.")
flags.DEFINE_float("sl_learning_rate", 0.01,
                   "Learning rate for avg-policy sl network.")
flags.DEFINE_string("optimizer_str", "sgd",
                    "Optimizer, choose from 'adam', 'sgd'.")
flags.DEFINE_string("loss_str", "mse",
                    "Loss function, choose from 'mse', 'huber'.")
flags.DEFINE_integer("update_target_network_every", 19200,
                     "Number of steps between DQN target network updates.")
flags.DEFINE_float("discount_factor", 1.0,
                   "Discount factor for future rewards.")
flags.DEFINE_integer("epsilon_decay_duration", int(20e6),
                     "Number of game steps over which epsilon is decayed.")
flags.DEFINE_float("epsilon_start", 0.06,
                   "Starting exploration parameter.")
flags.DEFINE_float("epsilon_end", 0.001,
                   "Final exploration parameter.")
flags.DEFINE_string("evaluation_metric", "nash_conv",
                    "Choose from 'exploitability', 'nash_conv'.")
flags.DEFINE_bool("use_checkpoints", True, "Save/load neural network weights.")
flags.DEFINE_string("checkpoint_dir", "/tmp/nfsp_test",
                    "Directory to save/load the agent.")


class NFSPPolicies(policy.Policy):
  """Joint policy to be evaluated."""

  def __init__(self, env, nfsp_policies, mode):
    game = env.game
    player_ids = list(range(FLAGS.num_players))
    super(NFSPPolicies, self).__init__(game, player_ids)
    self._policies = nfsp_policies
    self._mode = mode
    self._obs = {
        "info_state": [None] * FLAGS.num_players,
        "legal_actions": [None] * FLAGS.num_players
    }

  def action_probabilities(self, state, player_id=None):
    cur_player = state.current_player()
    legal_actions = state.legal_actions(cur_player)

    self._obs["current_player"] = cur_player
    self._obs["info_state"][cur_player] = (
        state.information_state_tensor(cur_player))
    self._obs["legal_actions"][cur_player] = legal_actions

    info_state = rl_environment.TimeStep(
        observations=self._obs, rewards=None, discounts=None, step_type=None)

    with self._policies[cur_player].temp_mode_as(self._mode):
      p = self._policies[cur_player].step(info_state, is_evaluation=True).probs
    prob_dict = {action: p[action] for action in legal_actions}
    return prob_dict


def main(unused_argv):
  logging.info("Loading %s", FLAGS.game_name)
  game = FLAGS.game_name
  num_players = FLAGS.num_players

  env_configs = {"players": num_players}
  env = rl_environment.Environment(game, **env_configs)
  info_state_size = env.observation_spec()["info_state"][0]
  num_actions = env.action_spec()["num_actions"]

  hidden_layers_sizes = [int(l) for l in FLAGS.hidden_layers_sizes]
  kwargs = {
      "replay_buffer_capacity": FLAGS.replay_buffer_capacity,
      "reservoir_buffer_capacity": FLAGS.reservoir_buffer_capacity,
      "min_buffer_size_to_learn": FLAGS.min_buffer_size_to_learn,
      "anticipatory_param": FLAGS.anticipatory_param,
      "batch_size": FLAGS.batch_size,
      "learn_every": FLAGS.learn_every,
      "rl_learning_rate": FLAGS.rl_learning_rate,
      "sl_learning_rate": FLAGS.sl_learning_rate,
      "optimizer_str": FLAGS.optimizer_str,
      "loss_str": FLAGS.loss_str,
      "update_target_network_every": FLAGS.update_target_network_every,
      "discount_factor": FLAGS.discount_factor,
      "epsilon_decay_duration": FLAGS.epsilon_decay_duration,
      "epsilon_start": FLAGS.epsilon_start,
      "epsilon_end": FLAGS.epsilon_end,
  }

  with tf.Session() as sess:
    # pylint: disable=g-complex-comprehension
    agents = [
        nfsp.NFSP(sess, idx, info_state_size, num_actions, hidden_layers_sizes,
                  **kwargs) for idx in range(num_players)
    ]
    joint_avg_policy = NFSPPolicies(env, agents, nfsp.MODE.average_policy)

    sess.run(tf.global_variables_initializer())

    if FLAGS.use_checkpoints:
      for agent in agents:
        if agent.has_checkpoint(FLAGS.checkpoint_dir):
          agent.restore(FLAGS.checkpoint_dir)

    for ep in range(FLAGS.num_train_episodes):
      if (ep + 1) % FLAGS.eval_every == 0:
        losses = [agent.loss for agent in agents]
        logging.info("Losses: %s", losses)
        if FLAGS.evaluation_metric == "exploitability":
          # Avg exploitability is implemented only for 2 players constant-sum
          # games, use nash_conv otherwise.
          expl = exploitability.exploitability(env.game, joint_avg_policy)
          logging.info("[%s] Exploitability AVG %s", ep + 1, expl)
        elif FLAGS.evaluation_metric == "nash_conv":
          nash_conv = exploitability.nash_conv(env.game, joint_avg_policy)
          logging.info("[%s] NashConv %s", ep + 1, nash_conv)
        else:
          raise ValueError(" ".join(("Invalid evaluation metric, choose from",
                                     "'exploitability', 'nash_conv'.")))
        if FLAGS.use_checkpoints:
          for agent in agents:
            agent.save(FLAGS.checkpoint_dir)
        logging.info("_____________________________________________")

      time_step = env.reset()
      while not time_step.last():
        player_id = time_step.observations["current_player"]
        agent_output = agents[player_id].step(time_step)
        action_list = [agent_output.action]
        time_step = env.step(action_list)

      # Episode is over, step all agents with final info state.
      for agent in agents:
        agent.step(time_step)


if __name__ == "__main__":
  app.run(main)



# coding=utf-8
# Copyright 2018 The Dopamine Authors and Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
# This file is a fork of the original Dopamine code incorporating changes for
# the multiplayer setting and the Hanabi Learning Environment.
#
"""The standard DQN replay memory.

This implementation is an out-of-graph replay memory + in-graph wrapper. It
supports vanilla n-step updates of the form typically found in the literature,
i.e. where rewards are accumulated for n steps and the intermediate trajectory
is not exposed to the agent. This does not allow, for example, performing
off-policy corrections.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import gzip
import math
import os
import pickle

import gin.tf
import numpy as np
import tensorflow as tf


# This constant determines how many iterations a checkpoint is kept for.
CHECKPOINT_DURATION = 4
MAX_SAMPLE_ATTEMPTS = 1000000


def invalid_range(cursor, replay_capacity, stack_size):
  """Returns an array with all the indices invalidated by cursor.

  It handles special cases in a circular buffer in the beginning and the end.

  Args:
    cursor: int, The position of the cursor.
    replay_capacity: int, The size of the replay memory.
    stack_size: int, The size of the stacks returned by the replay memory.

  Returns:
    `np.array` of size stack_size with the invalid indices.
  """
  assert cursor < replay_capacity
  return np.array(
      [(cursor - 1 + i) % replay_capacity for i in range(stack_size)])


class OutOfGraphReplayMemory(object):
  """A simple out of graph replay memory.

  Stores transitions (i.e. state, action, reward, next_state, terminal)
  efficiently when the states consist of stacks. The writing behaves like
  a FIFO buffer and the sampling is uniformly random.

  Attributes:
    add_count:  counter of how many transitions have been added.
    observations: `np.array`, circular buffer of observations.
    actions: `np.array`, circular buffer of actions.
    rewards: `np.array`, circular buffer of rewards.
    terminals: `np.array`, circular buffer of terminals.
    legal_actions: `np.array`, circular buffer of legal actions for hanabi.
    invalid_range: `np.array`, currently invalid indices.
  """

  def __init__(self, num_actions, observation_size, stack_size, replay_capacity,
               batch_size, update_horizon=1, gamma=1.0):
    """Data structure doing the heavy lifting.

    Args:
      num_actions: int, number of possible actions.
      observation_size: int, size of an input frame.
      stack_size: int, number of frames to use in state stack.
      replay_capacity: int, number of transitions to keep in memory.
      batch_size: int, batch size.
      update_horizon: int, length of update ('n' in n-step update).
      gamma: float, the discount factor.
    """
    self._observation_size = observation_size
    self._num_actions = num_actions
    self._replay_capacity = replay_capacity
    self._batch_size = batch_size
    self._stack_size = stack_size
    self._update_horizon = update_horizon
    self._gamma = gamma

    # When the horizon is > 1, we compute the sum of discounted rewards as a dot
    # product using the precomputed vector <gamma^0, gamma^1, ..., gamma^{n-1}>.
    self._cumulative_discount_vector = np.array(
        [math.pow(self._gamma, n) for n in range(update_horizon)],
        dtype=np.float32)

    # Create numpy arrays used to store sampled transitions.
    self.observations = np.empty(
        (replay_capacity, observation_size), dtype=np.uint8)
    self.actions = np.empty((replay_capacity), dtype=np.int32)
    self.rewards = np.empty((replay_capacity), dtype=np.float32)
    self.terminals = np.empty((replay_capacity), dtype=np.uint8)
    self.legal_actions = np.empty((replay_capacity, num_actions),
                                  dtype=np.float32)
    self.reset_state_batch_arrays(batch_size)
    self.add_count = np.array(0)

    self.invalid_range = np.zeros((self._stack_size))

  def add(self, observation, action, reward, terminal, legal_actions):
    """Adds a transition to the replay memory.

    Since the next_observation in the transition will be the observation added
    next there is no need to pass it.

    If the replay memory is at capacity the oldest transition will be discarded.

    Args:
      observation: `np.array` uint8, (observation_size).
      action: uint8, indicating the action in the transition.
      reward: float, indicating the reward received in the transition.
      terminal: uint8, acting as a boolean indicating whether the transition
                 was terminal (1) or not (0).
      legal_actions: Binary vector indicating legal actions (1 == legal).
    """
    if self.is_empty() or self.terminals[self.cursor() - 1] == 1:
      dummy_observation = np.zeros((self._observation_size))
      dummy_legal_actions = np.zeros((self._num_actions))
      for _ in range(self._stack_size - 1):
        self._add(dummy_observation, 0, 0, 0, dummy_legal_actions)
    self._add(observation, action, reward, terminal, legal_actions)

  def _add(self, observation, action, reward, terminal, legal_actions):
    cursor = self.cursor()
    self.observations[cursor] = observation
    self.actions[cursor] = action
    self.rewards[cursor] = reward
    self.terminals[cursor] = terminal
    self.legal_actions[cursor] = legal_actions
    self.add_count += 1
    self.invalid_range = invalid_range(self.cursor(), self._replay_capacity,
                                       self._stack_size)

  def is_empty(self):
    """Is the replay memory empty?"""
    return self.add_count == 0

  def is_full(self):
    """Is the replay memory full?"""
    return self.add_count >= self._replay_capacity

  def cursor(self):
    """Index to the location where the next transition will be written."""
    return self.add_count % self._replay_capacity

  def get_stack(self, array, index):
    """Returns the stack of array at the index.

    Args:
      array: `np.array`, to get the stack from.
      index: int, index to the first terminal in the stack to be returned.
    Returns:
      `Tensor` with shape (stack_size)
    """
    assert index >= 0
    assert index < self._replay_capacity
    if not self.is_full():
      assert index < self.cursor(), 'Index %i has not been added.' % index
      assert index >= self._stack_size - 1, ('Not enough elements to sample '
                                             'index %i' % index)
    # Fast slice read.
    if index >= self._stack_size - 1:
      stack = array[(index - self._stack_size + 1):(index + 1), ...]
    # Slow list read.
    else:
      indices = [(index - i) % self._replay_capacity
                 for i in reversed(range(self._stack_size))]
      stack = array[indices, ...]
    return stack

  def get_observation_stack(self, index):
    state = self.get_stack(self.observations, index)
    return np.transpose(state, [1, 0])

  def get_terminal_stack(self, index):
    return self.get_stack(self.terminals, index)

  def is_valid_transition(self, index):
    """Checks if the index contains a valid transition.

    The index range needs to be valid and it must not collide with the end of an
      episode.

    Args:
      index: int, index to the state in the transition. Note that next_state
        must also be valid.

    Returns:
      bool, True if transition is valid.

    """
    # Range checks
    if index < 0 or index >= self._replay_capacity:
      return False
    if not self.is_full():
      # The indices and next_indices must be smaller than the cursor.
      if index >= self.cursor() - self._update_horizon:
        return False
      # The first few indices contain the padding states of the first episode.
      if index < self._stack_size - 1:
        return False

    # Skip transitions that straddle the cursor.
    if index in set(self.invalid_range):
      return False

    # If there are terminal flags in any other frame other than the last one
    # the stack is not valid, so don't sample it.
    if self.get_terminal_stack(index)[:-1].any():
      return False

    return True

  def reset_state_batch_arrays(self, batch_size):
    self._next_state_batch = np.empty(
        (batch_size, self._observation_size, self._stack_size), dtype=np.uint8)
    self._state_batch = np.empty(
        (batch_size, self._observation_size, self._stack_size), dtype=np.uint8)

  def sample_index_batch(self, batch_size):
    """Returns a batch of valid indices.

    Args:
      batch_size: int, number of indices returned.

    Returns:
      list of batch_size, containing valid indices.

    Raises:
      Exception: If the batch was not constructed after maximum number of tries.
    """
    indices = []
    attempt_count = 0
    while len(indices) < batch_size and attempt_count < MAX_SAMPLE_ATTEMPTS:
      attempt_count += 1
      # index references the state and index + 1 points to next_state
      if self.is_full():
        index = np.random.randint(0, self._replay_capacity)
      else:
        # Can't start at 0 because the buffer is not yet circular
        index = np.random.randint(self._stack_size - 1, self.cursor() - 1)
      if self.is_valid_transition(index):
        indices.append(index)
    if len(indices) != batch_size:
      raise Exception('I tried %i times but only sampled %i valid transitions' %
                      (MAX_SAMPLE_ATTEMPTS, len(indices)))
    return indices

  def sample_transition_batch(self, batch_size=None, indices=None):
    """Returns a batch of transitions.

    Args:
      batch_size: int, number of transitions returned. If None the batch_size
        defined at init will be used.
      indices: list of ints. If not None, use the given indices instead of
        sampling them.

    Returns:
      Minibatch of transitions:  A tuple with elements state_batch,
        action_batch, reward_batch, next_state_batch and terminal_batch. The
        shape of state_batch and next_state_batch is (minibatch_size,
        observation_size, stack_size) and the rest of tensors have
        shape (minibatch_size)
    """
    if batch_size is None:
      batch_size = self._batch_size
    if batch_size != self._state_batch.shape[0]:
      self.reset_state_batch_arrays(batch_size)
    if not self.is_full():
      assert self.add_count >= batch_size, (
          'There is not enough to sample.'
          ' You need to call add at '
          'least %i (batch_size) times (currently it is %i)' % (batch_size,
                                                                self.add_count))
    if indices is None:
      indices = self.sample_index_batch(batch_size)
    assert len(indices) == batch_size

    action_batch = self.actions[indices]
    reward_batch = np.empty((batch_size), dtype=np.float32)
    terminal_batch = np.empty((batch_size), dtype=np.uint8)
    indices_batch = np.empty((batch_size), dtype=np.int32)
    next_legal_actions_batch = np.empty((batch_size, self._num_actions),
                                        dtype=np.float32)

    for batch_element, memory_index in enumerate(indices):
      indices_batch[batch_element] = memory_index

      self._state_batch[batch_element] = self.get_observation_stack(
          memory_index)

      # Compute indices in the replay memory up to n steps ahead.
      trajectory_indices = [(memory_index + j) % self._replay_capacity for
                            j in range(self._update_horizon)]

      # Determine if this trajectory segment contains a terminal state.
      terminals_in_trajectory = np.nonzero(self.terminals[trajectory_indices])
      if terminals_in_trajectory[0].size == 0:
        # If not, sum rewards along the trajectory, property discounted.
        terminal_batch[batch_element] = 0
        reward_batch[batch_element] = self._cumulative_discount_vector.dot(
            self.rewards[trajectory_indices])
      else:
        # Updates leading to a terminal state require a little more care, in
        # particular to avoid summing rewards past the end of the episode.
        terminal_batch[batch_element] = 1
        # Fetch smallest index corresponding to a terminal state.
        terminal_index = np.min(terminals_in_trajectory)

        truncated_discount_vector = (
            self._cumulative_discount_vector[0:terminal_index + 1])
        reward_batch[batch_element] = truncated_discount_vector.dot(
            self.rewards[trajectory_indices[0:terminal_index + 1]])

      bootstrap_state_index = (
          (memory_index + self._update_horizon) % self._replay_capacity)
      self._next_state_batch[batch_element] = (
          self.get_observation_stack(bootstrap_state_index))
      next_legal_actions_batch[batch_element] = (
          self.legal_actions[bootstrap_state_index])

    return (self._state_batch, action_batch, reward_batch,
            self._next_state_batch, terminal_batch, indices_batch,
            next_legal_actions_batch)

  def _generate_filename(self, checkpoint_dir, name, suffix):
    return os.path.join(checkpoint_dir, '{}_ckpt.{}.gz'.format(name, suffix))

  def save(self, checkpoint_dir, iteration_number):
    """Save the python replay memory attributes into a file.

    This method will save all the replay memory's state in a single file.

    Args:
      checkpoint_dir: str, directory where numpy checkpoint files should be
        saved.
      iteration_number: int, iteration_number to use as a suffix in naming numpy
        checkpoint files.
    """
    if not tf.gfile.Exists(checkpoint_dir):
      return
    for attr in self.__dict__:
      if not attr.startswith('_'):
        filename = self._generate_filename(checkpoint_dir, attr,
                                           iteration_number)
        with tf.gfile.Open(filename, 'wb') as f:
          with gzip.GzipFile(fileobj=f) as outfile:
            # Checkpoint numpy arrays directly with np.save to avoid excessive
            # memory usage. This is particularly important for the observations
            # data.
            if isinstance(self.__dict__[attr], np.ndarray):
              np.save(outfile, self.__dict__[attr], allow_pickle=False)
            else:
              pickle.dump(self.__dict__[attr], outfile)

      # After writing a checkpoint file, we garbage collect the checkpoint file
      # that is four versions old.
      stale_iteration_number = iteration_number - CHECKPOINT_DURATION
      if stale_iteration_number >= 0:
        stale_filename = self._generate_filename(checkpoint_dir, attr,
                                                 stale_iteration_number)
        try:
          tf.gfile.Remove(stale_filename)
        except tf.errors.NotFoundError:
          pass

  def load(self, checkpoint_dir, suffix):
    """Restores the object from bundle_dictionary and numpy checkpoints.

    Args:
      checkpoint_dir: str, directory where to read the numpy checkpointed files
        from.
      suffix: str, suffix to use in numpy checkpoint files.

    Raises:
      NotFoundError: if all expected files are not found in directory.
    """
    # We will first make sure we have all the necessary files available to avoid
    # loading a partially-specified (i.e. corrupted) replay buffer.
    for attr in self.__dict__:
      if attr.startswith('_'):
        continue
      filename = self._generate_filename(checkpoint_dir, attr, suffix)
      if not tf.gfile.Exists(filename):
        raise tf.errors.NotFoundError(None, None,
                                      'Missing file: {}'.format(filename))
    # If we've reached this point then we have verified that all expected files
    # are available.
    for attr in self.__dict__:
      if attr.startswith('_'):
        continue
      filename = self._generate_filename(checkpoint_dir, attr, suffix)
      with tf.gfile.Open(filename, 'rb') as f:
        with gzip.GzipFile(fileobj=f) as infile:
          if isinstance(self.__dict__[attr], np.ndarray):
            self.__dict__[attr] = np.load(infile, allow_pickle=False)
          else:
            self.__dict__[attr] = pickle.load(infile)


@gin.configurable(blacklist=['observation_size', 'stack_size'])
class WrappedReplayMemory(object):
  """In-graph wrapper for the python replay memory.

  Usage:
    To add a transition:  run the operation add_transition_op
                          (and feed all the placeholders in add_transition_ph).

    To sample a batch:    Construct operations that depend on any of the
                          sampling tensors. Every sess.run using any of these
                          tensors will sample a new transition.

    When using staging:   Need to prefetch the next batch with each train_op by
                          calling self.prefetch_batch.

                          Everytime this op is called a new transition batch
                          would be prefetched.

  Attributes:
    The following tensors are sampled randomly each sess.run:
      states actions rewards next_states terminals
    add_transition_op:    tf operation to add a transition to the replay memory.
    All the following placeholders need to be fed. add_obs_ph add_action_ph
    add_reward_ph add_terminal_ph
  """

  def __init__(self,
               num_actions,
               observation_size,
               stack_size,
               use_staging=True,
               replay_capacity=1000000,
               batch_size=32,
               update_horizon=1,
               gamma=1.0,
               wrapped_memory=None):
    """Initializes a graph wrapper for the python replay memory.

    Args:
      num_actions: int, number of possible actions.
      observation_size: int, size of an input frame.
      stack_size: int, number of frames to use in state stack.
      use_staging: bool, when True it would use a staging area to prefetch the
        next sampling batch.
      replay_capacity: int, number of transitions to keep in memory.
      batch_size: int.
      update_horizon: int, length of update ('n' in n-step update).
      gamma: int, the discount factor.
      wrapped_memory: The 'inner' memory data structure. Defaults to None, which
        creates the standard DQN replay memory.

    Raises:
      ValueError: If update_horizon is not positive.
      ValueError: If discount factor is not in [0, 1].
    """
    if replay_capacity < update_horizon + 1:
      raise ValueError('Update horizon (%i) should be significantly smaller '
                       'than replay capacity (%i).'
                       % (update_horizon, replay_capacity))
    if not update_horizon >= 1:
      raise ValueError('Update horizon must be positive.')
    if not 0.0 <= gamma <= 1.0:
      raise ValueError('Discount factor (gamma) must be in [0, 1].')

    # Allow subclasses to create self.memory.
    if wrapped_memory is not None:
      self.memory = wrapped_memory
    else:
      self.memory = OutOfGraphReplayMemory(
          num_actions, observation_size, stack_size,
          replay_capacity, batch_size, update_horizon, gamma)

    with tf.name_scope('replay'):
      with tf.name_scope('add_placeholders'):
        self.add_obs_ph = tf.placeholder(
            tf.uint8, [observation_size], name='add_obs_ph')
        self.add_action_ph = tf.placeholder(tf.int32, [], name='add_action_ph')
        self.add_reward_ph = tf.placeholder(
            tf.float32, [], name='add_reward_ph')
        self.add_terminal_ph = tf.placeholder(
            tf.uint8, [], name='add_terminal_ph')
        self.add_legal_actions_ph = tf.placeholder(
            tf.float32, [num_actions], name='add_legal_actions_ph')

      add_transition_ph = [
          self.add_obs_ph, self.add_action_ph, self.add_reward_ph,
          self.add_terminal_ph, self.add_legal_actions_ph
      ]

      with tf.device('/cpu:*'):
        self.add_transition_op = tf.py_func(
            self.memory.add, add_transition_ph, [], name='replay_add_py_func')

        self.transition = tf.py_func(
            self.memory.sample_transition_batch, [],
            [tf.uint8, tf.int32, tf.float32, tf.uint8, tf.uint8, tf.int32,
             tf.float32],
            name='replay_sample_py_func')

        if use_staging:
          # To hide the py_func latency use a staging area to pre-fetch the next
          # batch of transitions.
          (states, actions, rewards, next_states,
           terminals, indices, next_legal_actions) = self.transition
          # StagingArea requires all the shapes to be defined.
          states.set_shape([batch_size, observation_size, stack_size])
          actions.set_shape([batch_size])
          rewards.set_shape([batch_size])
          next_states.set_shape(
              [batch_size, observation_size, stack_size])
          terminals.set_shape([batch_size])
          indices.set_shape([batch_size])
          next_legal_actions.set_shape([batch_size, num_actions])

          # Create the staging area in CPU.
          prefetch_area = tf.contrib.staging.StagingArea(
              [tf.uint8, tf.int32, tf.float32, tf.uint8, tf.uint8, tf.int32,
               tf.float32])

          self.prefetch_batch = prefetch_area.put(
              (states, actions, rewards, next_states, terminals, indices,
               next_legal_actions))
        else:
          self.prefetch_batch = tf.no_op()

      if use_staging:
        # Get the sample_transition_batch in GPU. This would do the copy from
        # CPU to GPU.
        self.transition = prefetch_area.get()

      (self.states, self.actions, self.rewards, self.next_states,
       self.terminals, self.indices, self.next_legal_actions) = self.transition

      # Since these are py_func tensors, no information about their shape is
      # present. Setting the shape only for the necessary tensors
      self.states.set_shape([None, observation_size, stack_size])
      self.next_states.set_shape([None, observation_size, stack_size])

  def save(self, checkpoint_dir, iteration_number):
    """Save the underlying replay memory's contents in a file.

    Args:
      checkpoint_dir: str, directory from where to read the numpy checkpointed
        files.
      iteration_number: int, iteration_number to use as a suffix in naming
        numpy checkpoint files.
    """
    self.memory.save(checkpoint_dir, iteration_number)

  def load(self, checkpoint_dir, suffix):
    """Loads the replay memory's state from a saved file.

    Args:
      checkpoint_dir: str, directory from where to read the numpy checkpointed
        files.
      suffix: str, suffix to use in numpy checkpoint files.
    """
    self.memory.load(checkpoint_dir, suffix)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Python spiel example."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import logging
from absl import app
from absl import flags

import tensorflow.compat.v1 as tf
from open_spiel.python.algorithms import dqn
from open_spiel.python.algorithms import eva
from open_spiel.python.algorithms import policy_gradient
from open_spiel.python.environments import catch

FLAGS = flags.FLAGS

flags.DEFINE_integer("num_episodes", int(3e3), "Number of train episodes.")
flags.DEFINE_enum("algorithm", "dqn", ["dqn", "rpg", "qpg", "rm", "eva", "a2c"],
                  "Algorithms to run.")


def _eval_agent(env, agent, num_episodes):
  """Evaluates `agent` for `num_episodes`."""
  rewards = 0.0
  for _ in range(num_episodes):
    time_step = env.reset()
    episode_reward = 0
    while not time_step.last():
      agent_output = agent.step(time_step, is_evaluation=True)
      time_step = env.step([agent_output.action])
      episode_reward += time_step.rewards[0]
    rewards += episode_reward
  return rewards / num_episodes


def main_loop(unused_arg):
  """Trains a DQN agent in the catch environment."""
  env = catch.Environment()
  info_state_size = env.observation_spec()["info_state"][0]
  num_actions = env.action_spec()["num_actions"]

  train_episodes = FLAGS.num_episodes
  eval_interval = 50

  with tf.Session() as sess:
    if FLAGS.algorithm in {"rpg", "qpg", "rm", "a2c"}:
      agent = policy_gradient.PolicyGradient(
          sess,
          player_id=0,
          info_state_size=info_state_size,
          num_actions=num_actions,
          loss_str=FLAGS.algorithm,
          hidden_layers_sizes=[32, 32],
          batch_size=32,
          entropy_cost=0.001,
          critic_learning_rate=0.01,
          pi_learning_rate=0.01,
          num_critic_before_pi=5)
    elif FLAGS.algorithm == "dqn":
      agent = dqn.DQN(
          sess,
          player_id=0,
          state_representation_size=info_state_size,
          num_actions=num_actions,
          learning_rate=1e-3,
          replay_buffer_capacity=10000,
          hidden_layers_sizes=[32, 32],
          epsilon_decay_duration=2000,  # 10% total data
          update_target_network_every=250)
    elif FLAGS.algorithm == "eva":
      agent = eva.EVAAgent(
          sess,
          env,
          player_id=0,
          state_size=info_state_size,
          num_actions=num_actions,
          learning_rate=1e-3,
          trajectory_len=2,
          num_neighbours=2,
          mixing_parameter=0.95,
          memory_capacity=10000,
          dqn_hidden_layers=[32, 32],
          epsilon_decay_duration=2000,  # 10% total data
          update_target_network_every=250)
    else:
      raise ValueError("Algorithm not implemented!")

    sess.run(tf.global_variables_initializer())

    # Train agent
    for ep in range(train_episodes):
      time_step = env.reset()
      while not time_step.last():
        agent_output = agent.step(time_step)
        action_list = [agent_output.action]
        time_step = env.step(action_list)
      # Episode is over, step agent with final info state.
      agent.step(time_step)

      if ep and ep % eval_interval == 0:
        logging.info("-" * 80)
        logging.info("Episode %s", ep)
        logging.info("Loss: %s", agent.loss)
        avg_return = _eval_agent(env, agent, 100)
        logging.info("Avg return: %s", avg_return)


if __name__ == "__main__":
  app.run(main_loop)



# Copyright 2018 Google LLC
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    https://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Random Agent."""

import random
from hanabi_learning_environment.rl_env import Agent


class RandomAgent(Agent):
  """Agent that takes random legal actions."""

  def __init__(self, config, *args, **kwargs):
    """Initialize the agent."""
    self.config = config

  def act(self, observation):
    """Act based on an observation."""
    if observation['current_player_offset'] == 0:
      return random.choice(observation['legal_moves'])
    else:
      return None



# coding=utf-8
# Copyright 2018 The Dopamine Authors.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A lightweight logging mechanism for dopamine agents."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os
import pickle
import tensorflow as tf


CHECKPOINT_DURATION = 4


class Logger(object):
  """Class for maintaining a dictionary of data to log."""

  def __init__(self, logging_dir):
    """Initializes Logger.

    Args:
      logging_dir: str, Directory to which logs are written.
    """
    # Dict used by logger to store data.
    self.data = {}
    self._logging_enabled = True

    if not logging_dir:
      tf.logging.info('Logging directory not specified, will not log.')
      self._logging_enabled = False
      return
    # Try to create logging directory.
    try:
      tf.gfile.MakeDirs(logging_dir)
    except tf.errors.PermissionDeniedError:
      # If it already exists, ignore exception.
      pass
    if not tf.gfile.Exists(logging_dir):
      tf.logging.warning(
          'Could not create directory %s, logging will be disabled.',
          logging_dir)
      self._logging_enabled = False
      return
    self._logging_dir = logging_dir

  def __setitem__(self, key, value):
    """This method will set an entry at key with value in the dictionary.

    It will effectively overwrite any previous data at the same key.

    Args:
      key: str, indicating key where to write the entry.
      value: A python object to store.
    """
    if self._logging_enabled:
      self.data[key] = value

  def _generate_filename(self, filename_prefix, iteration_number):
    filename = '{}_{}'.format(filename_prefix, iteration_number)
    return os.path.join(self._logging_dir, filename)

  def log_to_file(self, filename_prefix, iteration_number):
    """Save the pickled dictionary to a file.

    Args:
      filename_prefix: str, name of the file to use (without iteration
        number).
      iteration_number: int, the iteration number, appended to the end of
        filename_prefix.
    """
    if not self._logging_enabled:
      tf.logging.warning('Logging is disabled.')
      return
    log_file = self._generate_filename(filename_prefix, iteration_number)
    with tf.gfile.GFile(log_file, 'w') as fout:
      pickle.dump(self.data, fout, protocol=pickle.HIGHEST_PROTOCOL)
    # After writing a checkpoint file, we garbage collect the log file
    # that is CHECKPOINT_DURATION versions old.
    stale_iteration_number = iteration_number - CHECKPOINT_DURATION
    if stale_iteration_number >= 0:
      stale_file = self._generate_filename(filename_prefix,
                                           stale_iteration_number)
      try:
        tf.gfile.Remove(stale_file)
      except tf.errors.NotFoundError:
        # Ignore if file not found.
        pass

  def is_logging_enabled(self):
    """Return if logging is enabled."""
    return self._logging_enabled



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""Tests for open_spiel.python.policy."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl.testing import absltest
from absl.testing import parameterized
import numpy as np

from open_spiel.python import policy
from open_spiel.python.algorithms import get_all_states
import pyspiel

_TIC_TAC_TOE_STATES = [
    {
        # ...
        # xoo
        # ..x
        "state": "3 4 8 5",
        "legal_actions": (0, 1, 2, 6, 7)
    },
    {
        # xo.
        # oxx
        # o..
        "state": "4 1 0 3 5 6",
        "legal_actions": (2, 7, 8)
    },
    {
        # ...
        # ...
        # ...
        "state": "",
        "legal_actions": (0, 1, 2, 3, 4, 5, 6, 7, 8)
    }
]


def test_policy_on_game(self, game, policy_object):
  """Checks the policy conforms to the conventions.

  Checks the Policy.action_probabilities contains only legal actions (but not
  necessarily all).
  Checks that the probabilities are positive and sum to 1.

  Args:
    self: The Test class. This methid targets as being used as a utility
      function to test policies.
    game: A `pyspiel.Game`, same as the one used in the policy.
    policy_object: A `policy.Policy` object on `game`. to test.
  """

  all_states = get_all_states.get_all_states(
      game,
      depth_limit=-1,
      include_terminals=False,
      include_chance_states=False,
      to_string=lambda s: s.information_state_string())

  for state in all_states.values():
    legal_actions = set(state.legal_actions())
    action_probabilities = policy_object.action_probabilities(state)

    for action in action_probabilities.keys():
      # We want a clearer error message to be able to debug.
      actions_missing = set(legal_actions) - set(action_probabilities.keys())
      illegal_actions = set(action_probabilities.keys()) - set(legal_actions)
      self.assertIn(
          action,
          legal_actions,
          msg="The action {} is present in the policy but is not a legal "
          "actions (these are {})\n"
          "Legal actions missing from policy: {}\n"
          "Illegal actions present in policy: {}".format(
              action, legal_actions, actions_missing, illegal_actions))

    sum_ = 0
    for prob in action_probabilities.values():
      sum_ += prob
      self.assertGreaterEqual(prob, 0)
    self.assertAlmostEqual(1, sum_)


_LEDUC_POKER = pyspiel.load_game("leduc_poker")


class CommonTest(parameterized.TestCase):

  @parameterized.parameters([
      policy.TabularPolicy(_LEDUC_POKER),
      policy.UniformRandomPolicy(_LEDUC_POKER),
      policy.FirstActionPolicy(_LEDUC_POKER),
  ])
  def test_policy_on_leduc(self, policy_object):
    test_policy_on_game(self, _LEDUC_POKER, policy_object)

  @parameterized.named_parameters([
      ("pyspiel.UniformRandom", pyspiel.UniformRandomPolicy(_LEDUC_POKER)),
  ])
  def test_cpp_policies_on_leduc(self, policy_object):
    test_policy_on_game(self, _LEDUC_POKER, policy_object)


class TabularTicTacToePolicyTest(parameterized.TestCase):

  # Enumerating all the states for tic tac toe is quite slow, so we do this
  # ony once.
  @classmethod
  def setUpClass(cls):
    super(TabularTicTacToePolicyTest, cls).setUpClass()
    cls.game = pyspiel.load_game("tic_tac_toe")
    cls.tabular_policy = policy.TabularPolicy(cls.game)

  def test_policy_shape(self):
    # Tic tac toe has 4520 decision states; ref
    # https://pubs.acs.org/doi/pdf/10.1021/acs.jcim.5b00324
    # There are 9 possible moves in the game (one per grid cell).
    # However, the TabularPolicy uses InformationState as keys, which in the
    # case of TicTacToe corresponds to the number of unique sequences (due to
    # perfect recall) requires by several algorithms, i.e. CFR.
    self.assertEqual(self.tabular_policy.action_probability_array.shape,
                     (294778, 9))

  def test_policy_attributes(self):
    # Verify the base class attributes of the policy
    self.assertEqual(self.tabular_policy.player_ids, [0, 1])

  @parameterized.parameters(*_TIC_TAC_TOE_STATES)
  def test_policy_at_state(self, state, legal_actions):
    index = self.tabular_policy.state_lookup[state]
    prob = 1 / len(legal_actions)
    np.testing.assert_array_equal(
        self.tabular_policy.action_probability_array[index],
        [prob if action in legal_actions else 0 for action in range(9)])

  @parameterized.parameters(*_TIC_TAC_TOE_STATES)
  def test_legal_actions_at_state(self, state, legal_actions):
    index = self.tabular_policy.state_lookup[state]
    np.testing.assert_array_equal(
        self.tabular_policy.legal_actions_mask[index],
        [1 if action in legal_actions else 0 for action in range(9)])

  def test_call_for_state(self):
    state = self.game.new_initial_state()
    state.apply_action(3)
    state.apply_action(4)
    state.apply_action(5)
    state.apply_action(6)
    state.apply_action(7)
    self.assertEqual(
        self.tabular_policy.action_probabilities(state), {
            0: 0.25,
            1: 0.25,
            2: 0.25,
            8: 0.25
        })

  def test_states_ordered_by_player(self):
    max_player0_index = max(
        self.tabular_policy.state_lookup[state]
        for state in self.tabular_policy.states_per_player[0])
    min_player1_index = min(
        self.tabular_policy.state_lookup[state]
        for state in self.tabular_policy.states_per_player[1])
    self.assertEqual(max_player0_index + 1, min_player1_index)

  def test_state_in(self):
    # Per state, we have 9 cells each with 3 possible states (o, x, empty)
    # Tic tac toe has 4520 decision states, but the tabular policy indexes by
    # InformationState, which leads to a larger number due to perfect recall
    self.assertEqual(self.tabular_policy.state_in.shape, (294778, 27))

  @parameterized.parameters(*_TIC_TAC_TOE_STATES)
  def test_policy_for_state_string(self, state, legal_actions):
    prob = 1 / len(legal_actions)
    np.testing.assert_array_equal(
        self.tabular_policy.policy_for_key(state),
        [prob if action in legal_actions else 0 for action in range(9)])


class TabularPolicyTest(parameterized.TestCase):

  def test_update_elementwise(self):
    game = pyspiel.load_game("kuhn_poker")
    tabular_policy = policy.TabularPolicy(game)
    state = "0pb"
    np.testing.assert_array_equal(
        tabular_policy.policy_for_key(state), [0.5, 0.5])
    tabular_policy.policy_for_key(state)[0] = 0.9
    tabular_policy.policy_for_key(state)[1] = 0.1
    np.testing.assert_array_equal(
        tabular_policy.policy_for_key(state), [0.9, 0.1])

  def test_update_slice(self):
    game = pyspiel.load_game("kuhn_poker")
    tabular_policy = policy.TabularPolicy(game)
    state = "2b"
    np.testing.assert_array_equal(
        tabular_policy.policy_for_key(state), [0.5, 0.5])
    tabular_policy.policy_for_key(state)[:] = [0.8, 0.2]
    np.testing.assert_array_equal(
        tabular_policy.policy_for_key(state), [0.8, 0.2])

  def test_state_ordering_is_deterministic(self):
    game = pyspiel.load_game("kuhn_poker")
    tabular_policy = policy.TabularPolicy(game)

    expected = {
        "0": 0,
        "0pb": 1,
        "1": 2,
        "1pb": 3,
        "2": 4,
        "2pb": 5,
        "1p": 6,
        "1b": 7,
        "2p": 8,
        "2b": 9,
        "0p": 10,
        "0b": 11,
    }
    self.assertEqual(expected, tabular_policy.state_lookup)

  def test_states(self):
    game = pyspiel.load_game("leduc_poker")
    tabular_policy = policy.TabularPolicy(game)
    i = 0
    for state in tabular_policy.states:
      self.assertEqual(i, tabular_policy.state_index(state))
      i += 1

    self.assertEqual(936, i)

  @parameterized.parameters((policy.FirstActionPolicy, "kuhn_poker"),
                            (policy.UniformRandomPolicy, "kuhn_poker"),
                            (policy.FirstActionPolicy, "leduc_poker"),
                            (policy.UniformRandomPolicy, "leduc_poker"))
  def test_can_turn_policy_into_tabular_policy(self, policy_class, game_name):
    game = pyspiel.load_game(game_name)
    realized_policy = policy_class(game)
    tabular_policy = realized_policy.to_tabular()
    for state in tabular_policy.states:
      self.assertEqual(
          realized_policy.action_probabilities(state),
          tabular_policy.action_probabilities(state))


class TabularRockPaperScissorsPolicyTest(absltest.TestCase):

  # Enumerating all the states for rock-paper-scissors is fast, but
  # we initialize only once for consistency with slower games.
  @classmethod
  def setUpClass(cls):
    super(TabularRockPaperScissorsPolicyTest, cls).setUpClass()
    game = pyspiel.load_game_as_turn_based("matrix_rps")
    cls.tabular_policy = policy.TabularPolicy(game)

  def test_policy_attributes(self):
    # Verify the base class attributes of the policy
    self.assertEqual(self.tabular_policy.player_ids, [0, 1])

  def test_tabular_policy(self):
    # Test that the tabular policy is uniform random in each state.
    np.testing.assert_array_equal(
        self.tabular_policy.action_probability_array,
        [[1 / 3, 1 / 3, 1 / 3], [1 / 3, 1 / 3, 1 / 3]])

  def test_states_lookup(self):
    # Test that there are two valid states, indexed as 0 and 1.
    game = pyspiel.load_game_as_turn_based("matrix_rps")
    state = game.new_initial_state()
    first_info_state = state.information_state_string()
    state.apply_action(state.legal_actions()[0])
    second_info_state = state.information_state_string()
    self.assertCountEqual(self.tabular_policy.state_lookup,
                          [first_info_state, second_info_state])
    self.assertCountEqual(self.tabular_policy.state_lookup.values(), [0, 1])

  def test_legal_actions_mask(self):
    # Test that all actions are valid in all states.
    np.testing.assert_array_equal(self.tabular_policy.legal_actions_mask,
                                  [[1, 1, 1], [1, 1, 1]])


class UniformRandomPolicyTest(absltest.TestCase):

  def test_policy_attributes(self):
    game = pyspiel.load_game("tiny_bridge_4p")
    uniform_random_policy = policy.UniformRandomPolicy(game)
    self.assertEqual(uniform_random_policy.player_ids, [0, 1, 2, 3])

  def test_policy_at_state(self):
    game = pyspiel.load_game("tic_tac_toe")
    uniform_random_policy = policy.UniformRandomPolicy(game)
    state = game.new_initial_state()
    state.apply_action(2)
    state.apply_action(4)
    state.apply_action(6)
    state.apply_action(8)
    self.assertEqual(
        uniform_random_policy.action_probabilities(state), {
            0: 0.2,
            1: 0.2,
            3: 0.2,
            5: 0.2,
            7: 0.2
        })

  def test_players_have_different_legal_actions(self):
    game = pyspiel.load_game("oshi_zumo")
    uniform_random_policy = policy.UniformRandomPolicy(game)
    state = game.new_initial_state()
    state.apply_actions([46, 49])
    # Started with 50 coins each, now have 4 and 1 respectively
    self.assertEqual(
        uniform_random_policy.action_probabilities(state, player_id=0), {
            0: 0.2,
            1: 0.2,
            2: 0.2,
            3: 0.2,
            4: 0.2
        })
    self.assertEqual(
        uniform_random_policy.action_probabilities(state, player_id=1), {
            0: 0.5,
            1: 0.5
        })


if __name__ == "__main__":
  absltest.main()



# coding=utf-8
# Copyright 2018 The Dopamine Authors and Google LLC.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
#
#
#
# This file is a fork of the original Dopamine code incorporating changes for
# the multiplayer setting and the Hanabi Learning Environment.
#
"""The entry point for running a Rainbow agent on Hanabi."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from absl import app
from absl import flags

from third_party.dopamine import logger

import run_experiment

FLAGS = flags.FLAGS

flags.DEFINE_multi_string(
    'gin_files', [],
    'List of paths to gin configuration files (e.g.'
    '"configs/hanabi_rainbow.gin").')
flags.DEFINE_multi_string(
    'gin_bindings', [],
    'Gin bindings to override the values set in the config files '
    '(e.g. "DQNAgent.epsilon_train=0.1").')

flags.DEFINE_string('base_dir', None,
                    'Base directory to host all required sub-directories.')

flags.DEFINE_string('checkpoint_dir', '',
                    'Directory where checkpoint files should be saved. If '
                    'empty, no checkpoints will be saved.')
flags.DEFINE_string('checkpoint_file_prefix', 'ckpt',
                    'Prefix to use for the checkpoint files.')
flags.DEFINE_string('logging_dir', '',
                    'Directory where experiment data will be saved. If empty '
                    'no checkpoints will be saved.')
flags.DEFINE_string('logging_file_prefix', 'log',
                    'Prefix to use for the log files.')


def launch_experiment():
  """Launches the experiment.

  Specifically:
  - Load the gin configs and bindings.
  - Initialize the Logger object.
  - Initialize the environment.
  - Initialize the observation stacker.
  - Initialize the agent.
  - Reload from the latest checkpoint, if available, and initialize the
    Checkpointer object.
  - Run the experiment.
  """
  if FLAGS.base_dir == None:
    raise ValueError('--base_dir is None: please provide a path for '
                     'logs and checkpoints.')

  run_experiment.load_gin_configs(FLAGS.gin_files, FLAGS.gin_bindings)
  experiment_logger = logger.Logger('{}/logs'.format(FLAGS.base_dir))

  environment = run_experiment.create_environment()
  obs_stacker = run_experiment.create_obs_stacker(environment)
  agent = run_experiment.create_agent(environment, obs_stacker)

  checkpoint_dir = '{}/checkpoints'.format(FLAGS.base_dir)
  start_iteration, experiment_checkpointer = (
      run_experiment.initialize_checkpointing(agent,
                                              experiment_logger,
                                              checkpoint_dir,
                                              FLAGS.checkpoint_file_prefix))

  run_experiment.run_experiment(agent, environment, start_iteration,
                                obs_stacker,
                                experiment_logger, experiment_checkpointer,
                                checkpoint_dir,
                                logging_file_prefix=FLAGS.logging_file_prefix)


def main(unused_argv):
  """This main function acts as a wrapper around a gin-configurable experiment.

  Args:
    unused_argv: Arguments (unused).
  """
  launch_experiment()

if __name__ == '__main__':
  app.run(main)



# Copyright 2019 DeepMind Technologies Ltd. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

# Lint as python3
"""Wraps third-party bridge bots to make them usable in OpenSpiel.

This code enables OpenSpiel interoperation for bots which implement the BlueChip
bridge protocol. This is widely used, e.g. in the World computer bridge
championships. For a rough outline of the protocol, see:
http://www.bluechipbridge.co.uk/protocol.htm

No formal specification is available. This implementation has been verified
to work correctly with WBridge5.

This bot controls a single player in the full game of bridge, including both the
bidding and play phase. It chooses its actions by invoking an external bot which
plays the full game of bridge. This means that each time the bot is asked for an
action, it sends up to three actions (one for each other player) to the external
bridge bot, and obtains an action in return.
"""

import re
import pyspiel

# Example session:
#
# Recv: Connecting "WBridge5" as ANYPL using protocol version 18
# Send: WEST ("WBridge5") seated
# Recv: WEST ready for teams
# Send: Teams: N/S "silent" E/W "bidders"
# Recv: WEST ready to start
# Send: Start of board
# Recv: WEST ready for deal
# Send: Board number 8. Dealer WEST. Neither vulnerable.
# Recv: WEST ready for cards
# Send: WEST's cards: S A T 9 5. H K 6 5. D Q J 8 7 6. C 7.
# Recv: WEST PASSES
# Recv: WEST ready for  NORTH's bid
# Send: EAST PASSES
# Recv: WEST ready for EAST's bid
# Send: EAST bids 1C
# Recv: WEST ready for  SOUTH's bid

# The game we support
GAME_STR = "bridge(use_double_dummy_result=False)"

# Template regular expressions for messages we receive
_CONNECT = 'Connecting "(?P<client_name>.*)" as ANYPL using protocol version 18'
_PLAYER_ACTION = ("(?P<seat>NORTH|SOUTH|EAST|WEST) "
                  "((?P<pass>PASSES)|(?P<dbl>DOUBLES)|(?P<rdbl>REDOUBLES)|bids "
                  "(?P<bid>[^ ]*)|(plays (?P<play>[23456789tjqka][cdhs])))"
                  "(?P<alert> Alert.)?")
_READY_FOR_OTHER = ("{seat} ready for "
                    "(((?P<other>[^']*)'s ((bid)|(card to trick \\d+)))"
                    "|(?P<dummy>dummy))")

# Templates for fixed messages we receive
_READY_FOR_TEAMS = "{seat} ready for teams"
_READY_TO_START = "{seat} ready to start"
_READY_FOR_DEAL = "{seat} ready for deal"
_READY_FOR_CARDS = "{seat} ready for cards"
_READY_FOR_BID = "{seat} ready for {other}'s bid"

# Templates for messages we send
_SEATED = '{seat} ("{client_name}") seated'
_TEAMS = 'Teams: N/S "north-south" E/W "east-west"'
_START_BOARD = "start of board"
_DEAL = "Board number {board}. Dealer NORTH. Neither vulnerable."
_CARDS = "{seat}'s cards: {hand}"
_OTHER_PLAYER_ACTION = "{player} {action}"
_PLAYER_TO_LEAD = "{seat} to lead"
_DUMMY_CARDS = "Dummy's cards: {}"

# BlueChip bridge protocol message constants
_SEATS = ["NORTH", "EAST", "SOUTH", "WEST"]
_TRUMP_SUIT = ["C", "D", "H", "S", "NT"]
_NUMBER_TRUMP_SUITS = len(_TRUMP_SUIT)
_SUIT = _TRUMP_SUIT[:4]
_NUMBER_SUITS = len(_SUIT)
_RANKS = ["2", "3", "4", "5", "6", "7", "8", "9", "T", "J", "Q", "K", "A"]
_LSUIT = [x.lower() for x in _SUIT]
_LRANKS = [x.lower() for x in _RANKS]

# OpenSpiel action ids
_ACTION_PASS = 52
_ACTION_DBL = 53
_ACTION_RDBL = 54
_ACTION_BID = 55  # First bid, i.e. 1C


def _bid_to_action(action_str):
  """Returns an OpenSpiel action id (an integer) from a BlueChip bid string."""
  level = int(action_str[0])
  trumps = _TRUMP_SUIT.index(action_str[1:])
  return _ACTION_BID + (level - 1) * _NUMBER_TRUMP_SUITS + trumps


def _play_to_action(action_str):
  """Returns an OpenSpiel action id (an integer) from a BlueChip card string."""
  rank = _LRANKS.index(action_str[0])
  suit = _LSUIT.index(action_str[1])
  return rank * _NUMBER_SUITS + suit


def _action_to_string(action):
  """Converts OpenSpiel action id (an integer) to a BlueChip action string.

  Args:
    action: an integer action id corresponding to a bid.

  Returns:
    A string in BlueChip format, e.g. 'PASSES' or 'bids 1H', or 'plays ck'.
  """
  if action == _ACTION_PASS:
    return "PASSES"
  elif action == _ACTION_DBL:
    return "DOUBLES"
  elif action == _ACTION_RDBL:
    return "REDOUBLES"
  elif action >= _ACTION_BID:
    level = str((action - _ACTION_BID) // _NUMBER_TRUMP_SUITS + 1)
    trumps = _TRUMP_SUIT[(action - _ACTION_BID) % _NUMBER_TRUMP_SUITS]
    return "bids " + level + trumps
  else:
    rank = action // _NUMBER_SUITS
    suit = action % _NUMBER_SUITS
    return "plays " + _LRANKS[rank] + _LSUIT[suit]


def _expect_regex(controller, regex):
  """Reads a line from the controller, parses it using the regular expression."""
  line = controller.read_line()
  match = re.match(regex, line)
  if not match:
    raise ValueError("Received '{}' which does not match regex '{}'".format(
        line, regex))
  return match.groupdict()


def _expect(controller, expected):
  """Reads a line from the controller, checks it matches expected line exactly."""
  line = controller.read_line()
  if expected != line:
    raise ValueError("Received '{}' but expected '{}'".format(line, expected))


def _hand_string(cards):
  """Returns the hand of the to-play player in the state in BlueChip format."""
  if len(cards) != 13:
    raise ValueError("Must have 13 cards")
  suits = [[] for _ in range(4)]
  for card in reversed(sorted(cards)):
    suit = card % 4
    rank = card // 4
    suits[suit].append(_RANKS[rank])
  for i in range(4):
    if suits[i]:
      suits[i] = _TRUMP_SUIT[i] + " " + " ".join(suits[i]) + "."
    else:
      suits[i] = _TRUMP_SUIT[i] + " -."
  return " ".join(suits)


def _connect(controller, seat):
  """Performs the initial handshake with a BlueChip bot."""
  client_name = _expect_regex(controller, _CONNECT)["client_name"]
  controller.send_line(_SEATED.format(seat=seat, client_name=client_name))
  _expect(controller, _READY_FOR_TEAMS.format(seat=seat))
  controller.send_line(_TEAMS)
  _expect(controller, _READY_TO_START.format(seat=seat))


def _new_deal(controller, seat, hand, board):
  """Informs a BlueChip bots that there is a new deal."""
  controller.send_line(_START_BOARD)
  _expect(controller, _READY_FOR_DEAL.format(seat=seat))
  controller.send_line(_DEAL.format(board=board))
  _expect(controller, _READY_FOR_CARDS.format(seat=seat))
  controller.send_line(_CARDS.format(seat=seat, hand=hand))


class BlueChipBridgeBot(pyspiel.Bot):
  """An OpenSpiel bot, wrapping a BlueChip bridge bot implementation."""

  def __init__(self, game, player_id, controller_factory):
    """Initializes an OpenSpiel `Bot` wrapping a BlueChip-compatible bot.

    Args:
      game: The OpenSpiel game object, should be an instance of
        `bridge(use_double_dummy_result=false)`.
      player_id: The id of the player the bot will act as, 0 = North (dealer), 1
        = East, 2 = South, 3 = West.
      controller_factory: Callable that returns new BlueChip controllers which
        must support methods `read_line` and `send_line`, and `terminate`.
    """
    pyspiel.Bot.__init__(self)
    if str(game) != GAME_STR:
      raise ValueError(f"BlueChipBridgeBot invoked with {game}")
    self._game = game
    self._player_id = player_id
    self._controller_factory = controller_factory
    self._seat = _SEATS[player_id]
    self._num_actions = 52
    self.dummy = None
    self.is_play_phase = False
    self.cards_played = 0
    self._board = 0
    self._state = self._game.new_initial_state()
    self._controller = None

  def player_id(self):
    return self._player_id

  def restart(self):
    """Indicates that we are starting a new episode."""
    # If we already have a fresh state, there is nothing to do.
    if not self._state.history():
      return
    self._num_actions = 52
    self.dummy = None
    self.is_play_phase = False
    self.cards_played = 0
    # We didn't see the end of the episode, so the external bot will still
    # be expecting it. If we can autoplay other people's actions to the end
    # (e.g. everyone passes or players play their last card), then do that.
    if not self._state.is_terminal():
      state = self._state.clone()
      while (not state.is_terminal()
             and state.current_player() != self._player_id):
        legal_actions = state.legal_actions()
        if _ACTION_PASS in legal_actions:
          state.apply(_ACTION_PASS)
        elif len(legal_actions) == 1:
          state.apply_action(legal_actions[0])
      if state.is_terminal():
        self.inform_state(state)
    # Otherwise, we will have to restart the external bot, because
    # the protocol makes no provision for this case.
    if not self._state.is_terminal():
      self._controller.terminate()
      self._controller = None
    self._state = self._game.new_initial_state()

  def _update_for_state(self):
    """Called for all non-chance nodes, whether or not we have to act."""
    # Get the actions in the game so far.
    actions = self._state.history()
    self.is_play_phase = (not self._state.is_terminal() and
                          max(self._state.legal_actions()) < 52)
    self.cards_played = sum(1 if a < 52 else 0 for a in actions) - 52

    # If this is the first time we've seen the deal, send our hand.
    if len(actions) == 52:
      self._board += 1
      _new_deal(self._controller, self._seat,
                _hand_string(actions[self._player_id:52:4]), self._board)

    # Send actions since last `step` call.
    for other_player_action in actions[self._num_actions:]:
      other = _expect_regex(self._controller,
                            _READY_FOR_OTHER.format(seat=self._seat))
      other_player = other["other"]
      if other_player == "Dummy":
        other_player = _SEATS[self.dummy]
      self._controller.send_line(
          _OTHER_PLAYER_ACTION.format(
              player=other_player,
              action=_action_to_string(other_player_action)))
    self._num_actions = len(actions)

    # If the opening lead has just been made, give the dummy.
    if self.is_play_phase and self.cards_played == 1:
      self.dummy = self._state.current_player() ^ 2
      if self._player_id != self.dummy:
        other = _expect_regex(self._controller,
                              _READY_FOR_OTHER.format(seat=self._seat))
        dummy_cards = _hand_string(actions[self.dummy:52:4])
        self._controller.send_line(_DUMMY_CARDS.format(dummy_cards))

    # If the episode is terminal, send (fake) timing info.
    if self._state.is_terminal():
      self._controller.send_line(
          "Timing - N/S : this board  [1:15],  total  [0:11:23].  "
          "E/W : this board  [1:18],  total  [0:10:23]"
      )
      self.dummy = None
      self.is_play_phase = False
      self.cards_played = 0

  def inform_action(self, state, player, action):
    del player, action
    self.inform_state(state)

  def inform_state(self, state):
    # Connect if we need to.
    if self._controller is None:
      self._controller = self._controller_factory()
      _connect(self._controller, self._seat)

    full_history = state.history()
    known_history = self._state.history()
    if full_history[:len(known_history)] != known_history:
      raise ValueError(
          "Supplied state is inconsistent with bot's internal state\n"
          f"Supplied state:\n{state}\n"
          f"Internal state:\n{self._state}\n")
    for action in full_history[len(known_history):]:
      self._state.apply_action(action)
      if not self._state.is_chance_node():
        self._update_for_state()

  def step(self, state):
    """Returns an action for the given state."""
    # Bring the external bot up-to-date.
    self.inform_state(state)

    # If we're on a new trick, tell the bot it is its turn.
    if self.is_play_phase and self.cards_played % 4 == 0:
      self._controller.send_line(_PLAYER_TO_LEAD.format(seat=self._seat))

    # Get our action from the bot.
    our_action = _expect_regex(self._controller, _PLAYER_ACTION)
    self._num_actions += 1
    if our_action["pass"]:
      return _ACTION_PASS
    elif our_action["dbl"]:
      return _ACTION_DBL
    elif our_action["rdbl"]:
      return _ACTION_RDBL
    elif our_action["bid"]:
      return _bid_to_action(our_action["bid"])
    elif our_action["play"]:
      return _play_to_action(our_action["play"])

  def terminate(self):
    self._controller.terminate()
    self._controller = None

